"""
Let the agent execute shell commands.

> [!WARNING]
> Using this tool will bypass the filesystem jail mechanism
"""

import subprocess
from typing import Annotated

from nerve.tools.utils import maybe_text

# for docs
EMOJI = "💻"


def shell(
    command: Annotated[str, "The shell command to execute"],
) -> str | bytes:
    """Execute a shell command on the local computer and return the output. Non interactive shell with a timeout of 30 seconds."""

    result = subprocess.run(command, shell=True, capture_output=True, timeout=30)

    raw_output = result.stdout or b""

    if result.returncode != 0:
        raw_output += b"\nEXIT CODE: " + str(result.returncode).encode("utf-8")

    if result.stderr:
        if result.returncode != 0:
            raw_output += b"\nERROR: " + result.stderr
        else:
            raw_output += b"\n" + result.stderr

    return maybe_text(raw_output)



import inspect
import typing as t
from typing import Annotated

from loguru import logger


def get_tool_schema(func: t.Callable[..., t.Any]) -> dict[str, t.Any]:
    signature = inspect.signature(func)
    docstring = inspect.getdoc(func) or ""

    if docstring == "":
        logger.debug(f"tool {func.__name__} has no docstring")

    type_hints = t.get_type_hints(func, include_extras=True)
    tool = {
        "type": "function",
        "function": {
            "name": func.__name__,
            "description": docstring,
            "parameters": {"type": "object", "properties": {}, "required": []},
        },
    }

    for param_name, param in signature.parameters.items():
        if param_name == "self":
            continue

        param_type = type_hints.get(param_name)
        if param_type is None:
            continue

        is_annotated = t.get_origin(param_type) is Annotated

        if is_annotated:
            args = t.get_args(param_type)
            base_type = args[0]
            description = args[1] if len(args) > 1 else ""
        else:
            base_type = param_type
            description = ""

        param_schema = process_type(base_type)

        if description:
            if isinstance(description, str):
                param_schema["description"] = description
            else:
                if hasattr(description, "description"):
                    param_schema["description"] = description.description

                if hasattr(description, "examples"):
                    param_schema["examples"] = description.examples

        tool["function"]["parameters"]["properties"][param_name] = param_schema  # type: ignore

        if param.default is param.empty:
            tool["function"]["parameters"]["required"].append(param_name)  # type: ignore

    return tool


def get_tool_response(response: t.Any) -> t.Any:
    response = response or ""
    if isinstance(response, str):
        # simple case, just text
        return response
    elif isinstance(response, bytes):
        # binary data, attempt decoding
        try:
            return response.decode("utf-8", errors="ignore")
        except UnicodeDecodeError:
            return str(response)

    elif isinstance(response, dict):
        # structured (vision), return as list
        return response
    else:
        logger.debug(f"unknown tool response type: {type(response)}")
        return get_tool_response(str(response))


def process_type(type_annotation: t.Any) -> dict[str, t.Any]:
    if type_annotation is str:
        return {"type": "string"}
    elif type_annotation is int:
        return {"type": "integer"}
    elif type_annotation is float:
        return {"type": "number"}
    elif type_annotation is bool:
        return {"type": "boolean"}
    elif type_annotation is list or t.get_origin(type_annotation) is list:
        item_type = t.get_args(type_annotation)[0] if t.get_args(type_annotation) else t.Any
        return {"type": "array", "items": process_type(item_type)}
    elif type_annotation is dict or t.get_origin(type_annotation) is dict:
        return {"type": "object"}

    if isinstance(type_annotation, type):
        return process_typed_dict(type_annotation)

    return {"type": "object"}


def process_typed_dict(typed_dict_class: type) -> dict[str, t.Any]:
    annotations = t.get_type_hints(typed_dict_class, include_extras=True)

    schema = {
        "type": "object",
        "properties": {},
        "required": [],
    }

    for field_name, field_type in annotations.items():
        is_annotated = t.get_origin(field_type) is Annotated

        if is_annotated:
            args = t.get_args(field_type)
            base_type = args[0]
            field_schema = process_type(base_type)

            for arg in args[1:]:
                if hasattr(arg, "description") and arg.description:
                    field_schema["description"] = arg.description
                if hasattr(arg, "examples") and arg.examples:
                    field_schema["examples"] = arg.examples
        else:
            field_schema = process_type(field_type)

        schema["properties"][field_name] = field_schema  # type: ignore

    return schema



import pathlib

import click
from loguru import logger

import nerve.runtime.state as state
from nerve.generation import Engine, WindowStrategy
from nerve.generation.conversation import FullHistoryStrategy
from nerve.generation.litellm import LiteLLMEngine
from nerve.models import Configuration, Tool, Usage
from nerve.runtime import Runtime


class Agent:
    def __init__(
        self,
        runtime: Runtime,
        configuration: Configuration,
        generation_engine: Engine,
        conv_window_strategy: WindowStrategy,
    ):
        # runtime data
        self.runtime = runtime
        # agent definition / configuration
        self.configuration = configuration
        # generation engine
        self.generation_engine = generation_engine
        # conversation window strategy
        self.conv_window_strategy = conv_window_strategy
        # user message to be added to the conversation at runtime if in interactive mode
        self._extra_message: str | None = None

        state.on_event("agent_created", {"agent": self})

    @classmethod
    def create(
        cls,
        generator: str,
        configuration: Configuration,
        start_state: dict[str, str] | None = None,
        window_strategy: WindowStrategy = FullHistoryStrategy(),
        working_dir: pathlib.Path = pathlib.Path.cwd(),
        name: str = "agent",
    ) -> "Agent":
        """
        Create an agent from a generator and configuration.

        Args:
            generator: The generator string to use if not set in the configuration.
            configuration: The configuration to use.
            start_state: Initial variables for the agent.
            window_strategy: How to handle conversation history.
            working_dir: The working directory to use.
            name: The name of the agent.
        """

        if start_state:
            state.update_variables(start_state)

        if configuration.defaults:
            state.set_defaults(configuration.defaults)

        # import tools.py by default if found
        tools_py_path = working_dir / "tools.py"
        if tools_py_path.exists():
            configuration.tools.append(
                Tool(
                    path=str(tools_py_path),
                )
            )

        configuration.generator = configuration.generator or generator

        runtime = Runtime.build(
            working_dir=working_dir,
            name=name,
            generator=configuration.generator,
            using=configuration.using,
            jail=configuration.jail,
            tools=configuration.tools,
        )

        return cls(
            runtime=runtime,
            configuration=configuration,
            generation_engine=LiteLLMEngine(configuration.generator, window_strategy, runtime.tools),
            conv_window_strategy=window_strategy,
        )

    @classmethod
    def create_from_file(
        cls,
        generator: str,
        config_file_path: pathlib.Path,
        window_strategy: WindowStrategy = FullHistoryStrategy(),
        start_state: dict[str, str] | None = None,
    ) -> "Agent":
        config = Configuration.from_path(config_file_path)
        if config.is_legacy:
            logger.error(f"legacy format detected, update to the 1.0.0 format: {config_file_path}")
            exit(1)

        stem = config_file_path.stem
        working_dir = (config_file_path if config_file_path.is_dir() else config_file_path.parent).absolute()
        return cls.create(
            generator,
            config,
            start_state,
            window_strategy,
            working_dir,
            stem if stem not in ("task", "agent") else working_dir.stem,
        )

    def _get_system_prompt(self) -> str | None:
        if not self.configuration.agent:
            return None

        raw = self.configuration.agent

        for name, value in state.get_knowledge().items():
            raw += f"\n\n## {name.capitalize()}\n\n{value}"

        return state.interpolate(raw)

    def _get_prompt(self) -> str:
        if not self.configuration.task:
            self.configuration.task = state.on_user_input_needed("task", "Describe the task: ")

        return state.interpolate(self.configuration.task)

    def add_extra_message(self, message: str) -> None:
        self._extra_message = message

    async def step(self) -> Usage:
        logger.debug(f"agent {self.runtime.name} step")

        if state.get_current_actor() != self:
            state.on_task_started(self)

        try:
            system_prompt = self._get_system_prompt()
            prompt = self._get_prompt()
            extra_tools = state.get_extra_tools()
            extra_message = None

            if self._extra_message:
                extra_message = self._extra_message
                self._extra_message = None

            logger.debug(f"system_prompt: {system_prompt}")
            logger.debug(f"prompt: {prompt}")
            logger.debug(f"extra_tools: {extra_tools}")

            state.on_event(
                "agent_step",
                {
                    "agent_name": self.runtime.name,
                    "generator": self.runtime.generator,
                    "system_prompt": system_prompt,
                    "prompt": prompt,
                },
            )

            usage = await self.generation_engine.step(system_prompt, prompt, extra_tools, extra_message)
            logger.debug(f"usage: {usage}")
            return usage
        except click.exceptions.MissingParameter as e:
            logger.error(e)
            exit(1)
        except Exception as e:
            state.on_event(
                "error",
                {
                    "agent_name": self.runtime.name,
                    "error": e,
                },
            )
            import traceback

            logger.error(f"Exception during agent step: {e}")
            logger.error(traceback.format_exc())
            quit()

    async def run(
        self,
        max_steps: int = 500,
        max_cost: float = 10.0,
        timeout: int | None = None,
        start_state: dict[str, str] | None = None,
    ) -> None:
        """
        Shortcut method to run a single agent without explicitly creating a Flow object.

        Runs the agent until one of the following conditions is met:

        - The task is completed.
        - The maximum number of steps is reached.
        - The maximum cost is reached.
        - The timeout is reached.

        Args:
            max_steps: The maximum number of steps to run.
            max_cost: The maximum cost in USD.
            timeout: The timeout in seconds.
            start_state: Initial variables for the agent.
        """

        # import here to avoid circular import
        from nerve.runtime.flow import Flow

        await Flow.build(
            actors=[self],
            max_steps=max_steps,
            max_cost=max_cost,
            timeout=timeout,
            start_state=start_state,
        ).run()



import asyncio
import json
import os
import pathlib
import threading
import typing as t

import click
import jinja2
from loguru import logger

from nerve.models import Mode, Status, Usage
from nerve.runtime.events import Event
from nerve.runtime.thread_pool import ThreadPool

# the current actor
_current_actor: t.Any | None = None
# total usage
_usage: Usage = Usage()
# event log
_events: list[Event] = []
# trace file
_trace_file: pathlib.Path | None = None
# working mode
_mode: Mode = Mode.AUTOMATIC
# the status of the active task
_task_status: Status = Status.RUNNING
# the reason for failed status
_reason: str | None = None
# variables
_variables: dict[str, t.Any] = {}
# defaults for variables
_defaults: dict[str, t.Any] = {}
# similar to variables but used by tools
_knowledge: dict[str, t.Any] = {}
# tools
_tools: dict[str, t.Callable[..., t.Any]] = {}
# extra tools defined at runtime
_extra_tools: dict[str, t.Callable[..., t.Any]] = {}
# listeners for events
_listeners: list[t.Callable[[Event], None]] = []
# thread pool for asynchronous event dispatching
_thread_pool: ThreadPool = ThreadPool()


def add_event_listener(listener: t.Callable[[Event], None]) -> None:
    """Add a listener function for events."""

    global _listeners
    _listeners.append(listener)


def set_trace_file(trace_file: pathlib.Path) -> None:
    """Enable recording of events to a file."""

    global _trace_file

    if trace_file.exists():
        logger.error(f"trace file {trace_file} already exists")
        exit(1)

    _trace_file = trace_file.absolute()
    logger.info(f"🔍 tracing to {_trace_file}")


class CustomJSONEncoder(json.JSONEncoder):
    def default(self, o: t.Any) -> t.Any:
        logger.debug("serializing", o)
        if hasattr(o, "model_dump"):
            return o.model_dump()
        elif hasattr(o, "__dict__"):
            return o.__dict__
        elif hasattr(o, "__str__"):
            return str(o)

        return super().default(o)


def wait_for_events_logs() -> None:
    """Wait for all events to be flushed."""

    global _thread_pool
    _thread_pool.wait_all()


def update_usage(usage: Usage) -> None:
    """Update the total usage."""

    global _usage
    _usage += usage


def get_usage() -> Usage:
    """Get the total usage."""

    global _usage
    return _usage


def on_event(name: str, data: t.Any | None = None) -> None:
    """Register an event."""

    global _events

    event = Event(name=name, data=data)
    _events.append(event)

    for listener in _listeners:
        # execute the listener in a separate thread in order to avoid blocking the main thread
        _task_id = _thread_pool.submit(listener, event)

    if _trace_file:
        with open(_trace_file, "a+t") as f:
            line = json.dumps(event.model_dump(), cls=CustomJSONEncoder)
            f.write(f"{line}\n")


def on_task_started(actor: t.Any) -> None:
    """Register a task start."""

    global _current_actor
    _current_actor = actor
    on_event("task_started", {"actor": actor})


def get_current_actor() -> t.Any:
    """Get the current actor."""

    global _current_actor
    return _current_actor


def on_before_tool_called(
    name: str,
    args: t.Any | None = None,
) -> None:
    """Register a tool call (before it is executed)."""

    on_event(
        "before_tool_called",
        {
            "name": name,
            "args": args,
        },
    )


def on_tool_called(
    started_at: float,
    finished_at: float,
    name: str,
    args: t.Any | None = None,
    result: t.Any | None = None,
    error: t.Any | None = None,
) -> None:
    """Register a tool call (after it is executed)."""

    on_event(
        "tool_called",
        {
            "started_at": started_at,
            "finished_at": finished_at,
            "name": name,
            "args": args,
            "result": result,
            "error": error,
        },
    )


def set_mode(new_mode: Mode) -> None:
    """Set the mode."""

    global _mode

    if _mode != new_mode:
        on_event("mode_change", {"from": _mode, "to": new_mode})

    _mode = new_mode


def get_mode() -> Mode:
    """Get the current mode."""

    global _mode
    return _mode


def is_interactive() -> bool:
    """Check if the current mode is interactive."""

    return get_mode() == Mode.INTERACTIVE


def set_defaults(defaults: dict[str, t.Any]) -> None:
    """Set the defaults for variables."""

    global _defaults
    _defaults = defaults


def is_active_task_done() -> bool:
    """Check if the active task is done."""

    return _task_status.is_done()


def set_task_complete(the_reason: str | None = None) -> None:
    """Set the task as complete."""

    global _task_status, _reason
    _task_status = Status.COMPLETED
    _reason = the_reason
    on_event(
        "task_complete", {"actor": _current_actor.runtime.name if _current_actor else "unknown", "reason": the_reason}
    )


def set_task_failed(the_reason: str) -> None:
    """Set the task as failed."""

    global _task_status, _reason
    _task_status = Status.FAILED
    _reason = the_reason

    on_event(
        "task_failed", {"actor": _current_actor.runtime.name if _current_actor else "unknown", "reason": the_reason}
    )


def on_max_steps_reached() -> None:
    """Set the task as failed due to max steps reached."""

    global _task_status, _reason
    if _task_status == Status.RUNNING:
        set_task_failed("max steps reached")


def on_max_cost_reached() -> None:
    """Set the task as failed due to max cost reached."""

    global _task_status, _reason
    if _task_status == Status.RUNNING:
        set_task_failed("max cost reached")


def on_timeout() -> None:
    """Set the task as failed due to timeout."""

    global _task_status, _reason
    if _task_status == Status.RUNNING:
        set_task_failed("timeout reached")


def as_dict() -> dict[str, t.Any]:
    """Get the current state as a dictionary."""

    return {
        "mode": _mode,
        "current_task": {
            "status": _task_status,
            "reason": _reason,
        },
        "variables": _variables,
        "knowledge": _knowledge,
    }


def set_tools(tools: dict[str, t.Callable[..., t.Any]]) -> None:
    """Set all tools."""

    global _tools
    _tools = tools

    logger.debug(f"tools: {_tools}")


def get_tools() -> dict[str, t.Callable[..., t.Any]]:
    """Get all tools."""

    global _tools
    return _tools


def get_extra_tools() -> dict[str, t.Callable[..., t.Any]]:
    """Get any extra tool registered at runtime."""

    global _extra_tools
    return _extra_tools


def set_extra_tool(tool: t.Callable[..., t.Any]) -> None:
    """Register an extra tool at runtime (used by anytool namespace)."""

    global _extra_tools
    _extra_tools[tool.__name__] = tool
    on_event("tool_created", {"name": tool.__name__, "code": tool.__code__.co_code})


def get_variable(key: str, default: t.Any = None) -> t.Any:
    """Get a variable."""

    global _variables
    return _variables.get(key, default)


def get_variables() -> dict[str, t.Any]:
    """Get all variables."""

    global _variables
    return _variables


def get_knowledge() -> dict[str, t.Any]:
    """Get the knowledge variable (a piece of information appearing in the system prompt)."""

    global _knowledge
    return _knowledge


def write_knowledge(key: str, value: t.Any) -> None:
    """Write a piece of knowledge that will be used in the system prompt."""

    global _knowledge

    on_event("knowledge_change", {"name": key, "from": _knowledge.get(key), "to": value})

    _knowledge[key] = value


def append_to_knowledge(key: str, value: t.Any) -> None:
    """Append a piece of knowledge to the existing knowledge."""

    global _knowledge
    if key not in _knowledge:
        write_knowledge(key, value)
    else:
        write_knowledge(key, _knowledge[key] + "\n" + value)


def clear_knowledge(key: str) -> None:
    """Remove a piece of knowledge."""

    global _knowledge
    if key in _knowledge:
        on_event("knowledge_change", {"name": key, "from": _knowledge[key], "to": None})
        del _knowledge[key]


def update_variables(update: dict[str, t.Any]) -> None:
    """Update variables."""

    global _variables
    for key, value in update.items():
        on_event("variable_change", {"name": key, "from": _variables.get(key), "to": value})
    _variables.update(update)


def reset() -> None:
    """Reset the state."""

    global _task_status, _reason, _knowledge
    _task_status = Status.RUNNING
    _reason = None
    _knowledge = {}


def on_user_input_needed(input_name: str, prompt: str) -> str:
    """Get user input."""

    # check if defined as environment variable
    if input_name in os.environ:
        return os.environ[input_name]

    elif input_name in _variables:
        return str(_variables[input_name])

    elif input_name in _defaults:
        # from the agent defaults
        return str(_defaults[input_name])

    elif is_interactive():
        # wait for all events to be logged
        wait_for_events_logs()
        # ask the user
        return input(prompt).strip()

    else:
        raise click.MissingParameter(
            param=click.Option(
                [f"--{input_name}"],
                type=str,
                required=True,
            ),
            message=f"Command line argument --{input_name} is required in non interactive mode.",
        )


def _create_jinja_env() -> jinja2.Environment:
    # we use this to catch undefined variables at runtime
    class OnUndefinedVariable(jinja2.Undefined):
        def __init__(self, *args: t.Any, **kwargs: t.Any) -> None:
            super().__init__(*args, **kwargs)

            global _variables

            undefined_name = self._undefined_name or ""

            logger.debug(f"undefined variable encountered: {undefined_name}")
            logger.debug(f"current variables: {_variables}")
            logger.debug(f"current defaults: {_defaults}")

            self.value = on_user_input_needed(undefined_name, f"Enter value for '{undefined_name}': ")

            # save to state
            update_variables({undefined_name or "": self.value or ""})

        def __str__(self) -> str:
            return self.value or "<UNDEFINED>"

    env = jinja2.Environment(undefined=OnUndefinedVariable)

    # allow prompts to call tools
    for name, tool_fn in _tools.items():
        # if the tool is async, wrap it in a sync function to make it callable from Jinja
        if asyncio.iscoroutinefunction(tool_fn):

            def make_sync_wrapper(fn: t.Callable[..., t.Any]) -> t.Callable[..., t.Any]:
                def sync_wrapper(*args: t.Any, **kwargs: t.Any) -> t.Any:
                    logger.debug(f"calling tool {fn.__name__} from template")
                    coro = fn(*args, **kwargs)
                    # NOTE: we use a list to store the result because nonlocal variables (like scalars) inside a nested
                    # function do not allow assignment unless explicitly declared nonlocal. However, mutable objects like
                    # lists can be modified within the nested function without extra declarations.
                    result_container = []

                    def we_need_an_async_loop_thread() -> None:
                        loop = asyncio.new_event_loop()
                        asyncio.set_event_loop(loop)
                        result_container.append(loop.run_until_complete(coro))

                    thread = threading.Thread(target=we_need_an_async_loop_thread)
                    thread.start()
                    thread.join()

                    return result_container[0]

                return sync_wrapper

            env.globals[name] = make_sync_wrapper(tool_fn)
        else:
            env.globals[name] = tool_fn

    return env


def interpolate(raw: str, extra: dict[str, t.Any] | None = None) -> str:
    """Interpolate the current state into a string."""

    env = _create_jinja_env()
    template = env.from_string(raw)
    context = _variables | (extra or {})

    return template.render(**context)



"""
Read-only access primitives to the local filesystem.
"""

import os
from typing import Annotated

from nerve.tools.utils import maybe_text, path_acl

# for docs
EMOJI = "📂"

# if set, the agent will only have access to these paths
jail: list[str] = []


def list_folder_contents(
    path: Annotated[str, "The path to the folder to list"],
) -> str:
    """List the contents of a folder on disk."""

    path_acl(path, jail)

    # The rationale here is that because of training data, models can
    # understand an "ls -la" output better than any custom output format
    # I could generate manually, so we just use the "ls -la" command to
    # list the contents of the folder.
    return os.popen(f"ls -la {path}").read()


def read_file(path: Annotated[str, "The path to the file to read"]) -> str | bytes:
    """Read the contents of a file from disk."""

    path_acl(path, jail)

    with open(path, "rb") as f:
        return maybe_text(f.read())



import os
from unittest.mock import MagicMock, patch

import pytest

from nerve.models import Mode
from nerve.runtime import state


class TestInterpolate:
    def setup_method(self) -> None:
        # Reset state before each test
        state.reset()
        state._variables = {}
        state._defaults = {}
        state._mode = Mode.AUTOMATIC
        state._tools = {}

    def test_basic_interpolation(self) -> None:
        # Setup
        state.update_variables({"name": "John", "age": 30})

        # Test
        result = state.interpolate("Hello, {{ name }}! You are {{ age }} years old.")

        # Verify
        assert result == "Hello, John! You are 30 years old."

    def test_interpolation_with_extra_context(self) -> None:
        # Setup
        state.update_variables({"name": "John"})
        extra = {"location": "New York"}

        # Test
        result = state.interpolate("{{ name }} is from {{ location }}.", extra)

        # Verify
        assert result == "John is from New York."

    def test_undefined_variable_in_automatic_mode(self) -> None:
        # Setup
        state._mode = Mode.AUTOMATIC
        state._defaults = {"city": "San Francisco"}

        # Test
        result = state.interpolate("Welcome to {{ city }}!")

        # Verify
        assert result == "Welcome to San Francisco!"

    def test_undefined_variable_from_environment(self) -> None:
        # Setup
        with patch.dict(os.environ, {"TEST_VAR": "test_value"}):
            # Test
            result = state.interpolate("Environment value: {{ TEST_VAR }}")

            # Verify
            assert result == "Environment value: test_value"
            # The variable should also be saved to state
            assert state._variables.get("TEST_VAR") == "test_value"

    @patch("builtins.input", return_value="Paris")
    def test_undefined_variable_in_interactive_mode(self, mock_input: MagicMock) -> None:
        # Setup
        state._mode = Mode.INTERACTIVE

        # Test
        result = state.interpolate("Welcome to {{ city }}!")

        # Verify
        assert result == "Welcome to Paris!"
        assert mock_input.called
        assert state._variables.get("city") == "Paris"

    def test_tool_call_in_template(self) -> None:
        # Setup
        mock_tool = MagicMock(return_value="tool result")
        state._tools = {"test_tool": mock_tool}

        # Test
        result = state.interpolate("Tool output: {{ test_tool() }}")

        # Verify
        assert result == "Tool output: tool result"
        assert mock_tool.called

    @patch("nerve.runtime.state.on_user_input_needed")
    def test_missing_variable_raises_exception(self, mock_input_needed: MagicMock) -> None:
        # Setup
        state._mode = Mode.AUTOMATIC
        mock_input_needed.side_effect = Exception("Missing parameter")

        # Test & Verify
        with pytest.raises(Exception):  # noqa: B017
            state.interpolate("Missing {{ variable }}")

    def test_complex_expression(self) -> None:
        # Setup
        state.update_variables({"numbers": [1, 2, 3, 4, 5]})

        # Test
        result = state.interpolate("Sum: {{ numbers|sum }}, Average: {{ (numbers|sum) / numbers|length }}")

        # Verify
        assert result == "Sum: 15, Average: 3.0"



import os
import queue
import threading
import time
import typing as t


class ThreadPool:
    def __init__(self, num_workers: int | None = None):
        self.tasks = queue.Queue()  # type: ignore
        self.results: dict[int, t.Any] = {}
        self.workers: list[threading.Thread] = []
        self.shutdown_flag = threading.Event()
        self.num_workers = num_workers or os.cpu_count() or 4
        self.task_counter = 0
        self.task_counter_lock = threading.Lock()

        # Create and start worker threads
        for _ in range(self.num_workers):
            worker = threading.Thread(target=self._worker_loop)
            worker.daemon = True
            worker.start()
            self.workers.append(worker)

    def _worker_loop(self) -> None:
        while not self.shutdown_flag.is_set():
            try:
                # Get task with timeout to periodically check shutdown flag
                task_id, func, args, kwargs = self.tasks.get(timeout=0.5)
                try:
                    result = func(*args, **kwargs)
                    self.results[task_id] = result
                except Exception as e:
                    self.results[task_id] = e
                finally:
                    self.tasks.task_done()
            except queue.Empty:
                continue

    def submit(self, func: t.Callable, *args, **kwargs) -> int:  # type: ignore
        """Submit a function to be executed by thread pool"""
        with self.task_counter_lock:
            self.task_counter += 1
            task_id = self.task_counter

        self.tasks.put((task_id, func, args, kwargs))
        return task_id

    def wait_for_task(self, task_id: int, timeout: float | None = None) -> t.Any:
        """Wait for a specific task to complete and return its result"""
        start_time = time.time()
        while task_id not in self.results:
            if timeout is not None and time.time() - start_time > timeout:
                raise TimeoutError(f"Task {task_id} did not complete within timeout")
            time.sleep(0.01)

        result = self.results[task_id]
        if isinstance(result, Exception):
            raise result
        return result

    def wait_all(self) -> None:
        """Wait for all tasks to complete"""
        self.tasks.join()

    def shutdown(self) -> None:
        """Signal workers to exit and wait for them to terminate"""
        self.shutdown_flag.set()
        for worker in self.workers:
            worker.join()



import pathlib
import typing as t
from enum import Enum

from pydantic import BaseModel, Field
from pydantic_yaml import parse_yaml_raw_as


class Mode(str, Enum):
    """
    Running mode.
    """

    INTERACTIVE = "interactive"
    AUTOMATIC = "automatic"


class Status(str, Enum):
    # used in interactive mode
    PAUSED = "paused"
    # the system is running (trying to solve the task)
    RUNNING = "running"
    # the task was completed successfully
    COMPLETED = "completed"
    # the task failed
    FAILED = "failed"

    def is_done(self) -> bool:
        return self in [Status.COMPLETED, Status.FAILED]


class Usage(BaseModel):
    cost: float | None = None
    prompt_tokens: int = 0
    completion_tokens: int = 0
    total_tokens: int = 0

    def __add__(self, other: "Usage") -> "Usage":
        return Usage(
            cost=(self.cost or 0) + (other.cost or 0),
            prompt_tokens=self.prompt_tokens + other.prompt_tokens,
            completion_tokens=self.completion_tokens + other.completion_tokens,
            total_tokens=self.total_tokens + other.total_tokens,
        )


class Tool(BaseModel):
    """
    A tool is a function that can be called by the agent.
    """

    class Argument(BaseModel):
        """
        An argument is a parameter of the tool.
        """

        name: str
        description: str
        example: str

    # if path is set, it'll be loaded from a python file
    path: str | None = None

    # yaml description of the tool
    name: str = ""
    description: str = ""
    arguments: list[Argument] = []
    complete_task: bool = False
    print: bool = False
    mime: str | None = None
    tool: str | None = None


class Configuration(BaseModel):
    """
    Configuration for an agent determining its "identity", task and capabilities.
    """

    # legacy field used to detect if the user is loading a legacy file
    system_prompt: str | None = Field(default=None, exclude=True)

    # optional generator
    generator: str | None = None
    # used for versioning the agents
    version: str = "1.0.0"
    # the system prompt, the agent identity
    agent: str | None = None
    # the main agent task
    task: str | None = None
    # optional default values for the initial state
    defaults: dict[str, t.Any] = {}
    # builtin namespaces
    using: list[str] = []
    # jail mechanism for each namespace
    jail: dict[str, list[str]] = {}
    # custom tooling
    tools: list[Tool | t.Callable[..., t.Any]] = []

    @staticmethod
    def is_agent_config(input_path: pathlib.Path) -> bool:
        try:
            _ = Configuration.from_path(input_path)
            return True
        except Exception:
            return False

    @classmethod
    def from_path(cls, input_path: pathlib.Path) -> "Configuration":
        if input_path.is_dir():
            for option in ("task.yml", "agent.yml"):
                sub_path = input_path / option
                if sub_path.exists():
                    input_path = sub_path
                    break

        with open(input_path) as f:
            return parse_yaml_raw_as(cls, f.read())

    @classmethod
    def from_yml(cls, config_yml: str) -> "Configuration":
        return parse_yaml_raw_as(cls, config_yml)

    @property
    def is_legacy(self) -> bool:
        return self.system_prompt is not None


class Workflow(BaseModel):
    """
    A workflow is a collection of agents that are executed sequentially.
    """

    class Actor(BaseModel):
        """
        An actor is an agent that is part of the workflow.
        """

        generator: str

    name: str
    description: str
    actors: dict[str, Actor]

    @staticmethod
    def is_workflow(input_path: pathlib.Path) -> bool:
        try:
            _ = Workflow.from_path(input_path)
            return True
        except Exception:
            return False

    @classmethod
    def from_path(
        cls,
        input_path: pathlib.Path,
    ) -> "Workflow":
        if input_path.is_dir():
            input_path = input_path / "workflow.yml"

        with open(input_path) as f:
            return parse_yaml_raw_as(cls, f.read())


__all__ = ["Mode", "Status", "Tool", "Configuration", "Workflow"]



import asyncio
import pathlib
import typing as t

import typer
from loguru import logger

import nerve
from nerve.cli.defaults import (
    DEFAULT_AGENTS_LOAD_PATH,
    DEFAULT_CONVERSATION_STRATEGY,
    DEFAULT_GENERATOR,
    DEFAULT_MAX_COST,
    DEFAULT_MAX_STEPS,
    DEFAULT_TIMEOUT,
)
from nerve.generation import WindowStrategy, conversation
from nerve.models import Configuration, Mode, Workflow
from nerve.runtime import logging, state
from nerve.runtime.agent import Agent
from nerve.runtime.flow import Flow

cli = typer.Typer(
    no_args_is_help=True,
    pretty_exceptions_enable=False,
    context_settings={"help_option_names": ["-h", "--help"]},
)


@cli.command(
    context_settings={"allow_extra_args": True, "ignore_unknown_options": True, "help_option_names": ["-h", "--help"]},
    help="Execute an agent or a workflow.",
)
def run(
    ctx: typer.Context,
    input_path: t.Annotated[
        pathlib.Path,
        typer.Argument(help="Agent or workflow to execute"),
    ] = pathlib.Path("."),
    generator: t.Annotated[
        str,
        typer.Option("--generator", "-g", help="If the agent generator field is not set, use this generator."),
    ] = DEFAULT_GENERATOR,
    conversation_strategy: t.Annotated[
        str,
        typer.Option("--conversation", "-c", help="Conversation strategy to use"),
    ] = DEFAULT_CONVERSATION_STRATEGY,
    interactive: t.Annotated[
        bool,
        typer.Option("--interactive", "-i", help="Interactive mode"),
    ] = False,
    debug: t.Annotated[
        bool,
        typer.Option("--debug", help="Enable debug logging"),
    ] = False,
    quiet: t.Annotated[
        bool,
        typer.Option("--quiet", "-q", help="Quiet mode"),
    ] = False,
    max_steps: t.Annotated[
        int,
        typer.Option("--max-steps", "-s", help="Maximum number of steps"),
    ] = DEFAULT_MAX_STEPS,
    max_cost: t.Annotated[
        float,
        typer.Option(
            "--max-cost", help="If cost information is available, stop when the cost exceeds this value in USD."
        ),
    ] = DEFAULT_MAX_COST,
    timeout: t.Annotated[
        int | None,
        typer.Option("--timeout", "-t", help="Timeout in seconds"),
    ] = DEFAULT_TIMEOUT,
    log_path: t.Annotated[
        pathlib.Path | None,
        typer.Option("--log", help="Log to a file."),
    ] = None,
    trace: t.Annotated[
        pathlib.Path | None,
        typer.Option("--trace", help="Save the final state to a file."),
    ] = None,
) -> None:
    logging.init(log_path, level="DEBUG" if debug else "SUCCESS" if quiet else "INFO")
    logger.info(f"🧠 nerve v{nerve.__version__}")

    asyncio.run(
        execute_flow(
            input_path,
            generator,
            # convert the conversation strategy string to a valid enum
            conversation.strategy_from_string(conversation_strategy),
            ctx.args,
            max_steps,
            max_cost,
            timeout,
            interactive,
            trace,
        )
    )


def _get_start_state(args: list[str]) -> dict[str, str]:
    # any unknown argument will populate the start_state
    start_state = {}
    for i in range(0, len(args), 2):
        key = args[i].removeprefix("--").removeprefix("-").replace("-", "_")
        value = args[i + 1] if i + 1 < len(args) else ""
        start_state[key] = value
    return start_state


def _resolve_input_path(input_path: pathlib.Path) -> pathlib.Path:
    # check if input_path exists
    if not input_path.exists():
        if not input_path.is_absolute():
            # check if it exists as part of the $HOME/.nerve/agents directory
            in_home = DEFAULT_AGENTS_LOAD_PATH / input_path
            if in_home.exists():
                input_path = in_home

            in_home_with_yaml = in_home.with_suffix(".yml")
            if in_home_with_yaml.exists():
                input_path = in_home_with_yaml

        if not input_path.exists():
            logger.error(f"path '{input_path}' does not exist")
            raise typer.Abort()

    return input_path


async def execute_flow(
    input_path: pathlib.Path,
    generator: str,
    conv_window_strategy: WindowStrategy,
    start_state_args: list[str],
    max_steps: int = 100,
    max_cost: float = 10.0,
    timeout: int | None = None,
    interactive: bool = False,
    trace: pathlib.Path | None = None,
) -> None:
    if trace:
        state.set_trace_file(trace)

    if interactive:
        state.set_mode(Mode.INTERACTIVE)

    # check if input_path exists
    input_path = _resolve_input_path(input_path)

    # make variables available to the runtime
    start_state = _get_start_state(start_state_args)
    state.update_variables(start_state)

    # check if input_path is a workflow or single agent
    if Workflow.is_workflow(input_path):
        # full workflow
        flow = Flow.from_path(
            input_path,
            window_strategy=conv_window_strategy,
            max_steps=max_steps,
            max_cost=max_cost,
            timeout=timeout,
            start_state=start_state,
        )

    elif Configuration.is_agent_config(input_path):
        # single agent
        flow = Flow.build(
            actors=[Agent.create_from_file(generator, input_path, conv_window_strategy)],
            max_steps=max_steps,
            max_cost=max_cost,
            timeout=timeout,
            start_state=start_state,
        )

    else:
        logger.error(f"path '{input_path}' is not a valid workflow or agent configuration")
        raise typer.Abort()

    await flow.run()



import typing as t

from loguru import logger

from nerve.generation import WindowStrategy


class FullHistoryStrategy(WindowStrategy):
    """
    This strategy returns the full history of the conversation.
    """

    async def get_window(self, history: list[dict[str, t.Any]]) -> list[dict[str, t.Any]]:
        return history

    def __str__(self) -> str:
        return "<full history>"


class SlidingWindowStrategy(WindowStrategy):
    """
    This strategy returns a sliding window of the last N messages.
    """

    def __init__(self, window_size: int = 10) -> None:
        self.window_size = window_size

    async def get_window(self, history: list[dict[str, t.Any]]) -> list[dict[str, t.Any]]:
        history_size = len(history)
        if history_size <= self.window_size:
            return history

        for msg in history:
            logger.debug(" * " + str(msg))
        logger.debug("---")

        # NOTE: This is a "best effort" sliding window strategy.
        #
        # In theory we could just get the last N messages, however for each
        # tool call from the AI, we need to add its tool call response or the
        # API errors with:
        #
        #   Invalid parameter: messages with role 'tool' must be a response
        #   to a preceeding message with 'tool_calls'.
        window = history[-self.window_size :]
        orphans = []
        for item in window:
            if item.get("role") == "tool":
                tool_call_id = item.get("tool_call_id", "")
                found = False
                for other in window:
                    # quick and dirty way to check if the tool call id is in the other item
                    if item != other and tool_call_id in str(other):
                        found = True
                        break

                if not found:
                    orphans.append(item)

        if orphans:
            window = [item for item in window if item not in orphans]

        return window

    def __str__(self) -> str:
        return f"<sliding window of size {self.window_size}>"


class StrippedWindowStrategy(WindowStrategy):
    """
    This strategy returns a sliding window of the last N messages, but instead of removing
    earlier messages it will replace the content of the messages with a <stripped> placeholder.
    """

    def __init__(self, window_size: int = 10) -> None:
        self.window_size = window_size

    async def get_window(self, history: list[dict[str, t.Any]]) -> list[dict[str, t.Any]]:
        history_size = len(history)
        if history_size <= self.window_size:
            return history

        window = []
        # calculate which messages should be stripped (all except the last window_size messages)
        messages_to_strip = max(0, history_size - self.window_size)

        for current_index, msg in enumerate(history):
            # create a copy of the message to avoid modifying the original
            msg = msg.copy()

            # strip content for messages outside the window (earlier messages)
            if current_index < messages_to_strip:
                if msg.get("role") == "tool":
                    # For tool messages, preserve tool_call_id but strip content
                    msg["content"] = "<stripped tool response>"
                elif "content" in msg:
                    msg["content"] = "<stripped content>"

            window.append(msg)

        return window

    def __str__(self) -> str:
        return f"<stripping window of size {self.window_size}>"


def strategy_from_string(strategy: str) -> WindowStrategy:
    if strategy == "full":
        return FullHistoryStrategy()
    elif strategy.startswith("strip-") and strategy.split("-")[1].isdigit():
        return StrippedWindowStrategy(int(strategy.split("-")[1]))
    elif strategy.isdigit():
        return SlidingWindowStrategy(int(strategy))
    else:
        raise ValueError(f"Invalid conversation strategy: {strategy}")



import asyncio
import typing as t
import unittest
from unittest.mock import patch

from nerve.generation.conversation import (
    FullHistoryStrategy,
    SlidingWindowStrategy,
    StrippedWindowStrategy,
    strategy_from_string,
)


class TestFullHistoryStrategy(unittest.TestCase):
    def test_get_window_returns_full_history(self) -> None:
        strategy = FullHistoryStrategy()
        history = [{"role": "user", "content": "Hello"}, {"role": "assistant", "content": "Hi"}]

        result = asyncio.run(strategy.get_window(history))

        self.assertEqual(result, history)
        self.assertEqual(str(strategy), "<full history>")


class TestSlidingWindowStrategy(unittest.TestCase):
    def test_get_window_returns_full_history_when_smaller_than_window(self) -> None:
        strategy = SlidingWindowStrategy(window_size=5)
        history = [{"role": "user", "content": "Hello"}, {"role": "assistant", "content": "Hi"}]

        result = asyncio.run(strategy.get_window(history))

        self.assertEqual(result, history)

    def test_get_window_returns_window_sized_history(self) -> None:
        strategy = SlidingWindowStrategy(window_size=2)
        history = [
            {"role": "user", "content": "First"},
            {"role": "user", "content": "Second"},
            {"role": "user", "content": "Third"},
            {"role": "user", "content": "Fourth"},
        ]

        result = asyncio.run(strategy.get_window(history))

        self.assertEqual(len(result), 2)
        self.assertEqual(result, history[2:])

    @patch("nerve.generation.conversation.logger.debug")
    def test_get_window_logs_messages(self, mock_debug: unittest.mock.Mock) -> None:
        strategy = SlidingWindowStrategy(window_size=2)
        history = [
            {"role": "user", "content": "First"},
            {"role": "assistant", "content": "First response"},
            {"role": "user", "content": "Second"},
        ]

        asyncio.run(strategy.get_window(history))

        # Verify logger was called
        self.assertTrue(mock_debug.called)

    def test_str_representation(self) -> None:
        strategy = SlidingWindowStrategy(window_size=10)
        self.assertEqual(str(strategy), "<sliding window of size 10>")


class TestStrippedWindowStrategy(unittest.TestCase):
    def test_get_window_returns_full_history_when_smaller_than_window(self) -> None:
        strategy = StrippedWindowStrategy(window_size=5)
        history = [{"role": "user", "content": "Hello"}, {"role": "assistant", "content": "Hi"}]

        result = asyncio.run(strategy.get_window(history))

        self.assertEqual(result, history)

    def test_get_window_strips_content_outside_window(self) -> None:
        strategy = StrippedWindowStrategy(window_size=2)
        history = [
            {"role": "user", "content": "First"},
            {"role": "user", "content": "Second"},
            {"role": "user", "content": "Third"},
            {"role": "user", "content": "Fourth"},
        ]

        result = asyncio.run(strategy.get_window(history))

        # All messages should be present
        self.assertEqual(len(result), 4)

        # First two messages should be stripped
        self.assertEqual(result[0]["content"], "<stripped content>")
        self.assertEqual(result[1]["content"], "<stripped content>")

        # Last two messages should be preserved
        self.assertEqual(result[2]["content"], "Third")
        self.assertEqual(result[3]["content"], "Fourth")

    def test_get_window_strips_tool_messages(self) -> None:
        strategy = StrippedWindowStrategy(window_size=1)
        history = [
            {"role": "user", "content": "First"},
            {"role": "tool", "tool_call_id": "call_123", "content": "Tool response"},
            {"role": "user", "content": "Third"},
        ]

        result = asyncio.run(strategy.get_window(history))

        # First message should be stripped
        self.assertEqual(result[0]["content"], "<stripped content>")

        # Tool message should be stripped but preserve tool_call_id
        self.assertEqual(result[1]["content"], "<stripped tool response>")
        self.assertEqual(result[1]["tool_call_id"], "call_123")

        # Last message should be preserved
        self.assertEqual(result[2]["content"], "Third")

    def test_str_representation(self) -> None:
        strategy = StrippedWindowStrategy(window_size=10)
        self.assertEqual(str(strategy), "<stripping window of size 10>")

    def test_strip_strategy_from_string(self) -> None:
        strategy = strategy_from_string("strip-8")
        self.assertIsInstance(strategy, StrippedWindowStrategy)
        self.assertEqual(t.cast(StrippedWindowStrategy, strategy).window_size, 8)


class TestStrategyFromString(unittest.TestCase):
    def test_full_strategy(self) -> None:
        strategy = strategy_from_string("full")
        self.assertIsInstance(strategy, FullHistoryStrategy)

    def test_sliding_window_strategy(self) -> None:
        strategy = strategy_from_string("5")
        self.assertIsInstance(strategy, SlidingWindowStrategy)
        self.assertEqual(t.cast(SlidingWindowStrategy, strategy).window_size, 5)

    def test_stripped_window_strategy(self) -> None:
        strategy = strategy_from_string("strip-2")
        self.assertIsInstance(strategy, StrippedWindowStrategy)
        self.assertEqual(t.cast(StrippedWindowStrategy, strategy).window_size, 2)

    def test_invalid_strategy(self) -> None:
        with self.assertRaises(ValueError):
            strategy_from_string("invalid")



import pathlib
import sys
import typing as t

from loguru import logger
from termcolor import colored

from nerve.runtime import state
from nerve.runtime.events import Event


def init(log_path: pathlib.Path | None = None, level: str = "INFO", litellm_debug: bool = False) -> None:
    """
    Initialize the logging system.

    Args:
        log_path: The path to the log file.
        level: The log level to use.
        litellm_debug: Whether to enable litellm debug logging.
    """
    format = "{time} {message}"

    if level != "DEBUG":
        logger.remove()
        logger.add(
            sys.stdout,
            colorize=True,
            format=format,
            level=level,
        )

    if litellm_debug:
        import litellm

        litellm._turn_on_debug()  # type: ignore

    if log_path:
        logger.add(log_path, format=format, level=level)

    state.add_event_listener(log_event_to_terminal)


# a dynamic object that will access dictionary attributes
class DictWrapper:
    def __init__(self, data_dict: dict[str, t.Any]) -> None:
        self._data = data_dict

    def __getattr__(self, name: str) -> t.Any:
        if name in self._data:
            if isinstance(self._data[name], dict):
                return DictWrapper(self._data[name])

            elif isinstance(self._data[name], list):
                return [DictWrapper(item) for item in self._data[name]]

            return self._data[name]

        return "unknown"

    def __str__(self) -> str:
        return str(self._data)


def log_event_to_terminal(event: Event) -> None:
    data = event.data or {}
    if event.name == "flow_started":
        if isinstance(data["flow"], dict):
            data["flow"] = DictWrapper(data["flow"])

        max_steps = data["flow"].max_steps
        max_cost = data["flow"].max_cost
        timeout = data["flow"].timeout
        timeout = f"{timeout}s timeout" if timeout else "no timeout"
        conv_window_strategy = data["flow"].actors[0].conv_window_strategy
        logger.info(f"🚀 max steps: {max_steps} | max cost: {max_cost:.2f}$ | {timeout} | {conv_window_strategy}")

    elif event.name == "agent_created":
        if isinstance(data["agent"], dict):
            data["agent"] = DictWrapper(data["agent"])

        generator = data["agent"].runtime.generator
        name = data["agent"].runtime.name
        version = data["agent"].configuration.version
        tools = len(data["agent"].runtime.tools)
        logger.info(f"🤖 {generator} | {name} v{version} with {tools} tools")

    elif event.name == "before_tool_called":
        # avoid logging twice
        if data["name"] in ("task_complete", "task_failed"):
            return

        args_str = ", ".join([colored(v, "yellow") for v in data["args"].values()])
        name = colored(data["name"], attrs=["bold"])
        logger.info(f"🛠️  {name}({args_str})")

    elif event.name == "tool_called":
        # avoid logging twice
        if data["name"] in ("task_complete", "task_failed"):
            return

        elapsed_time = data["finished_at"] - data["started_at"]
        if data["result"] is None:
            ret = "<none>"
        else:
            ret = f"{type(data['result'])} ({len(str(data['result']))} bytes)"

        logger.info(colored(f" ↳ {data['name']} -> {ret} in {elapsed_time:.4f}s", "dark_grey"))

    elif event.name == "task_complete":
        logger.info(colored(f"✅ task {data['actor']} completed", "green", attrs=["bold"]))

    elif event.name == "task_failed":
        logger.error(colored(f"❌ task {data['actor']} failed: {data['reason']}", "red", attrs=["bold"]))

    elif event.name == "tool_created":
        logger.info(f"🧰 registered tool: {data['name']}")

    elif event.name == "unknown_tool":
        logger.warning(f"❌ model called unknown tool: {data['tool_name']}")

    elif event.name == "tool_error":
        logger.error(f"❌ error executing {data['tool_name']}({data['args']}): {data['error']}")

    elif event.name == "flow_complete":
        if isinstance(data["usage"], dict):
            data["usage"] = DictWrapper(data["usage"])

        if data["usage"].total_tokens > 0:
            logger.info(f"⚙️  flow complete in {data['steps']} steps ({data['usage']})")
        else:
            logger.info(f"⚙️  flow complete in {data['steps']} steps")

    elif event.name == "text_response":
        logger.info(f"💬 {colored(data['response'], 'black', 'on_white')}")

    elif event.name == "step_started":
        # avoid logging step usage in interactive mode
        if state.is_interactive():
            return

        if isinstance(data["usage"], dict):
            data["usage"] = DictWrapper(data["usage"])

        if data["usage"].total_tokens > 0:
            logger.info(f"📊 [step {data['step']}] [usage {data['usage']}]")
        else:
            logger.info(f"📊 [step {data['step']}]")

    elif event.name in (
        "task_started",
        "agent_step",
        "step_complete",
        "variable_change",
        "knowledge_change",
        "mode_change",
    ):
        pass
    else:
        logger.info(f"unknown event: {event}")



import asyncio
import json
import os
import typing as t
from abc import ABC, abstractmethod

from loguru import logger

from nerve.models import Usage
from nerve.runtime import state
from nerve.tools.protocol import get_tool_response, get_tool_schema


class WindowStrategy(ABC):
    @abstractmethod
    async def get_window(self, history: list[dict[str, t.Any]]) -> list[dict[str, t.Any]]:
        pass

    @abstractmethod
    def __str__(self) -> str:
        pass


class Engine(ABC):
    def __init__(
        self,
        generator_id: str,
        window_strategy: WindowStrategy,
        tools: list[t.Callable[..., t.Any]] | None = None,
    ):
        self.generator_id = generator_id
        self.generator_params: dict[str, t.Any] = {}

        self._parse_generator_params()

        self.history: list[dict[str, t.Any]] = []
        self.window_strategy = window_strategy

        self.tools = {fn.__name__: fn for fn in (tools or [])}
        self.tools_schemas = []
        for tool_name, tool_fn in self.tools.items():
            if not tool_fn.__doc__:
                raise ValueError(f"Tool {tool_name} has no docstring")
            self.tools_schemas.append(get_tool_schema(tool_fn))

    def _parse_generator_params(self) -> None:
        if "?" in self.generator_id:
            # split generator_id by '?' and parse the right part as query parameters
            base_id, query_params = self.generator_id.split("?", 1)
            self.generator_id = base_id
            # parse as parameters
            for param in query_params.split("&"):
                key, value = param.split("=", 1)
                # try to convert value to float if it looks like a number
                try:
                    if "." in value:
                        self.generator_params[key] = float(value)
                    elif value.isdigit():
                        self.generator_params[key] = int(value)
                    else:
                        self.generator_params[key] = value
                except Exception:
                    # if conversion fails, keep as string
                    self.generator_params[key] = value

        # set API base from parameters or environment variable
        if "api_base" in self.generator_params:
            self.api_base = self.generator_params["api_base"]
            del self.generator_params["api_base"]
        elif "GENERATOR_API_BASE" in os.environ:
            self.api_base = os.environ["GENERATOR_API_BASE"]
        else:
            self.api_base = None

    def _get_extended_tooling_schema(self, extra_tools: dict[str, t.Callable[..., t.Any]]) -> list[dict[str, t.Any]]:
        tools_schemas = self.tools_schemas.copy()
        extra_schemas = []

        for tool_name, tool_fn in extra_tools.items():
            if tool_name in self.tools:
                raise ValueError(f"Tool {tool_name} already exists")
            else:
                logger.debug(f"adding extra tool: {tool_name} / {tool_fn.__name__}")
                extra_schemas.append(get_tool_schema(tool_fn))

        tools_schemas.extend(extra_schemas)

        logger.trace(tools_schemas)
        logger.debug(f"{len(tools_schemas)} tools")

        return tools_schemas

    def _get_text_response(self, content: str) -> dict[str, t.Any]:
        state.on_event(
            "text_response",
            {
                "generator": self.generator_id,
                "response": content,
            },
        )

        return {
            "role": "user",
            "content": "None of the tools were used, interact with the user by executing the existing tools.",
        }

    def _get_unknown_tool_response(self, tool_call_id: str, tool_name: str) -> dict[str, t.Any]:
        state.on_event(
            "unknown_tool",
            {
                "generator": self.generator_id,
                "tool_name": tool_name,
            },
        )

        return {
            "tool_call_id": tool_call_id,
            "role": "tool",
            "content": f"The tool {tool_name} is not available.",
        }

    async def _get_tool_response(
        self, tool_call_id: str, tool_name: str, tool_fn: t.Callable[..., t.Any], tool_args: dict[str, t.Any]
    ) -> list[dict[str, t.Any]]:
        logger.debug(f"calling tool: {tool_name} with args: {tool_args}")
        try:
            tool_response = tool_fn(**tool_args)
            # check if the tool function returned a coroutine
            if asyncio.iscoroutine(tool_response):
                tool_response = await tool_response
        except Exception as e:
            state.on_event(
                "tool_error",
                {
                    "generator": self.generator_id,
                    "tool_name": tool_name,
                    "args": tool_args,
                    "error": e,
                },
            )
            tool_response = f"ERROR while executing tool {tool_name}: {e}"

        generated_response = get_tool_response(tool_response)
        if isinstance(generated_response, str):
            return [
                {
                    "tool_call_id": tool_call_id,
                    "role": "tool",
                    "name": tool_name,
                    "content": generated_response,
                }
            ]
        else:
            """
            Handle:

                Invalid 'messages[3]'. Image URLs are only allowed for messages with role 'user', but this message with role 'tool' contains an image URL.", 'type': 'invalid_request_error', 'param': 'messages[3]', 'code': 'invalid_value'}}

            And:

                An assistant message with 'tool_calls' must be followed by tool messages responding to each 'tool_call_id'.
            """
            return [
                {
                    "tool_call_id": tool_call_id,
                    "role": "tool",
                    "name": tool_name,
                    "content": "",
                },
                {
                    "role": "user",
                    "content": [
                        {
                            "type": "text",
                            "text": f"{tool_name} returned the following response:",
                        },
                        generated_response,
                    ],
                },
            ]

    async def _process_tool_call(
        self, call_id: str, tool_name: str, args: str | dict[str, t.Any], extra_tools: dict[str, t.Callable[..., t.Any]]
    ) -> list[dict[str, t.Any]]:
        # resolve tool
        tool_fn = self.tools.get(tool_name, extra_tools.get(tool_name, None))
        if tool_fn is None:
            # unknown tool
            return [self._get_unknown_tool_response(call_id, tool_name)]
        else:
            tool_call_args = json.loads(args) if isinstance(args, str) else args
            # execute tool and collect response
            return await self._get_tool_response(call_id, tool_name, tool_fn, tool_call_args)

    @abstractmethod
    async def step(
        self,
        system_prompt: str | None,
        user_prompt: str,
        extra_tools: dict[str, t.Callable[..., t.Any]] | None = None,
        extra_message: str | None = None,
    ) -> Usage:
        pass



from loguru import logger
from termcolor import colored

from nerve.runtime import state
from nerve.runtime.agent import Agent


class Shell:
    def __init__(self) -> None:
        self._keep_going: bool = False
        # TODO: can we build this by inspecting the class methods?
        self.commands = [
            {
                "name": "quit",
                "description": "Stop the execution and exit.",
                "aliases": ["q", "exit"],
                "action": self._handle_quit,
            },
            {
                "name": "continue",
                "description": "Continue the execution until completion.",
                "aliases": ["c", "cont", "go"],
                "action": self._handle_continue,
            },
            {
                "name": "step",
                "description": "Execute a single step.",
                "aliases": ["", "s"],
                "action": self._handle_step,
            },
            {
                "name": "view",
                "description": "Inspect the current state.",
                "aliases": ["v"],
                "action": self._handle_view,
            },
            {
                "name": "help",
                "description": "Show this help message.",
                "aliases": ["h"],
                "action": self._handle_help,
            },
        ]

    async def _handle_quit(self) -> None:
        logger.warning("👋 quitting")
        exit(0)

    async def _handle_continue(self) -> bool:
        logger.info("▶️  running until completion ...")
        self._keep_going = True
        return True

    async def _handle_step(self) -> bool:
        logger.debug("▶️  single step")
        self._keep_going = False
        return True

    async def _handle_view(self) -> bool:
        usage = state.get_usage()
        print(colored("usage", "white", attrs=["bold"]))
        print(f"  prompt tokens: {usage.prompt_tokens}")
        print(f"  completion tokens: {usage.completion_tokens}")
        print(f"  total tokens: {usage.total_tokens}")
        print(f"  cost: {usage.cost} $")
        print()

        agent = state.get_current_actor()
        print(colored("agent", "white", attrs=["bold"]))
        print(f"  name: {agent.runtime.name}")
        print(f"  model: {agent.configuration.generator}")
        print(f"  conversation strategy: {agent.conv_window_strategy}")
        print()

        vars = state.get_variables()
        if vars:
            print(colored("variables", "white", attrs=["bold"]))
            for k, v in vars.items():
                print(f"  {colored(k, 'green')} = {v}")
            print()

        knowledge = state.get_knowledge()
        if knowledge:
            print(colored("knowledge", "white", attrs=["bold"]))
            for k, v in knowledge.items():
                print(f"  {colored(k, 'green')} = {v}")
            print()

        tools = state.get_tools()
        if tools:
            print(colored("tools", "white", attrs=["bold"]))
            for k in tools.keys():
                print(f"  {colored(k, 'magenta')}")
            print()

        extra_tools = state.get_extra_tools()
        if extra_tools:
            print(colored("extra tools", "white", attrs=["bold"]))
            for k in extra_tools.keys():
                print(f"  {colored(k, 'magenta')}")
            print()

        return False

    async def _handle_help(self) -> bool:
        for cmd in self.commands:
            print(f"  {colored(cmd['name'], 'yellow')} ({', '.join(cmd['aliases'])}): {cmd['description']}")  # type: ignore

        print("\nAnything else will be interpreted and used as a chat message for the current agent.")

        return False

    async def _handle_command(self, command: str) -> tuple[bool, bool]:
        command = command.lower()
        for cmd in self.commands:
            if command in cmd["aliases"] or command == cmd["name"]:  # type: ignore
                done = await cmd["action"]()  # type: ignore
                return True, done

        return False, False

    async def _show_prompt(self, actor: Agent) -> str:
        print()
        prompt = f"{colored(actor.runtime.name, 'yellow')} > "
        command = input(prompt).strip()
        print()

        return command

    async def interact_if_needed(self, actor: Agent) -> None:
        if not state.is_interactive() or self._keep_going:
            return

        # wait for all events to be logged
        state.wait_for_events_logs()

        # show the prompt and get the command
        command = await self._show_prompt(actor)

        # handle the command
        is_handled, done = await self._handle_command(command)
        if not is_handled:
            # anything else is chat
            logger.debug(f"unhandled command: {command}")
            # add the command to the conversation
            actor.add_extra_message(command)
        elif not done:
            # keep running the interaction shell
            await self.interact_if_needed(actor)



import os
import pathlib

DEFAULT_GENERATOR: str = os.getenv("NERVE_GENERATOR", "openai/gpt-4o-mini")
DEFAULT_MAX_STEPS: int = int(os.getenv("NERVE_MAX_STEPS", 100))
DEFAULT_MAX_COST: float = float(os.getenv("NERVE_MAX_COST", 10.0))
DEFAULT_TIMEOUT: int | None = int(os.getenv("NERVE_TIMEOUT", 0)) or None
DEFAULT_CONVERSATION_STRATEGY: str = os.getenv("NERVE_CONVERSATION_STRATEGY", "full")

DEFAULT_NERVE_HOME: pathlib.Path = pathlib.Path.home() / ".nerve"

DEFAULT_AGENTS_LOAD_PATH: pathlib.Path = DEFAULT_NERVE_HOME / "agents"
DEFAULT_PROMPTS_LOAD_PATH: pathlib.Path = DEFAULT_NERVE_HOME / "prompts"

DEFAULT_AGENT_PATH: pathlib.Path = pathlib.Path("agent.yml")
DEFAULT_AGENT_SYSTEM_PROMPT: str = "You are a helpful assistant."
DEFAULT_AGENT_TASK: str = "Make an HTTP request to {{{{ url }}}}"
DEFAULT_AGENT_TOOLS: list[str] = ["shell", "task"]



import os
import tempfile
import unittest
from pathlib import Path

from nerve.tools.namespaces import filesystem


class TestFilesystem(unittest.TestCase):
    def setUp(self) -> None:
        # Create temporary directory for testing
        self.temp_dir = tempfile.TemporaryDirectory()
        self.test_dir = Path(self.temp_dir.name)

        # Create test files and subdirectories
        self.test_file = self.test_dir / "test_file.txt"
        self.test_file.write_text("test content")

        self.test_subdir = self.test_dir / "subdir"
        self.test_subdir.mkdir()

        self.test_subfile = self.test_subdir / "subfile.txt"
        self.test_subfile.write_text("subfile content")

        # Reset jail between tests
        filesystem.jail = []

    def tearDown(self) -> None:
        self.temp_dir.cleanup()

    def test_list_folder_contents_no_jail(self) -> None:
        # Test without jail restrictions
        result = filesystem.list_folder_contents(str(self.test_dir))
        self.assertIn("test_file.txt", result)
        self.assertIn("subdir", result)

    def test_read_file_no_jail(self) -> None:
        # Test without jail restrictions
        result = filesystem.read_file(str(self.test_file))
        self.assertEqual(result, "test content")

    def test_list_folder_contents_with_jail_allowed(self) -> None:
        # Set jail to allow only the test directory
        filesystem.jail = [str(self.test_dir)]

        # Should work for the test directory
        result = filesystem.list_folder_contents(str(self.test_dir))
        self.assertIn("test_file.txt", result)

        # Should work for subdirectories
        result = filesystem.list_folder_contents(str(self.test_subdir))
        self.assertIn("subfile.txt", result)

    def test_read_file_with_jail_allowed(self) -> None:
        # Set jail to allow only the test directory
        filesystem.jail = [str(self.test_dir)]

        # Should work for files in the test directory
        result = filesystem.read_file(str(self.test_file))
        self.assertEqual(result, "test content")

        # Should work for files in subdirectories
        result = filesystem.read_file(str(self.test_subfile))
        self.assertEqual(result, "subfile content")

    def test_list_folder_contents_with_jail_denied(self) -> None:
        # Create another temporary directory outside the jail
        with tempfile.TemporaryDirectory() as outside_dir:
            # Set jail to allow only the test directory
            filesystem.jail = [str(self.test_dir)]

            # Should raise ValueError for directories outside the jail
            with self.assertRaises(ValueError):
                filesystem.list_folder_contents(outside_dir)

    def test_read_file_with_jail_denied(self) -> None:
        # Create a file outside the jail
        with tempfile.NamedTemporaryFile(mode="w+") as outside_file:
            outside_file.write("outside content")
            outside_file.flush()

            # Set jail to allow only the test directory
            filesystem.jail = [str(self.test_dir)]

            # Should raise ValueError for files outside the jail
            with self.assertRaises(ValueError):
                filesystem.read_file(outside_file.name)

    def test_multiple_jail_paths(self) -> None:
        # Create another temporary directory
        with tempfile.TemporaryDirectory() as second_dir:
            second_path = Path(second_dir)
            second_file = second_path / "second_file.txt"
            second_file.write_text("second content")

            # Set jail to allow both directories
            filesystem.jail = [str(self.test_dir), second_dir]

            # Should work for both directories
            result1 = filesystem.read_file(str(self.test_file))
            self.assertEqual(result1, "test content")

            result2 = filesystem.read_file(str(second_file))
            self.assertEqual(result2, "second content")

    def test_path_allowed_with_symlinks(self) -> None:
        # Create a directory outside the jail
        with tempfile.TemporaryDirectory() as outside_dir:
            outside_path = Path(outside_dir)
            outside_file = outside_path / "outside_file.txt"
            outside_file.write_text("outside content")

            # Create a symlink inside the jail pointing outside
            symlink_path = self.test_dir / "symlink_to_outside"
            os.symlink(outside_dir, str(symlink_path))

            # Set jail to allow only the test directory
            filesystem.jail = [str(self.test_dir)]

            # Should raise ValueError for symlinks that resolve outside the jail
            with self.assertRaises(ValueError):
                filesystem.list_folder_contents(str(symlink_path))

            with self.assertRaises(ValueError):
                filesystem.read_file(str(symlink_path / "outside_file.txt"))

    def test_read_file_binary(self) -> None:
        # Create a binary file with non-UTF8 content
        binary_file = self.test_dir / "binary_file"
        binary_content = bytes([0xFF, 0xFE, 0xFD])  # Invalid UTF-8 sequence
        binary_file.write_bytes(binary_content)

        # Set jail to allow the test directory
        filesystem.jail = [str(self.test_dir)]

        # Should return bytes for binary content
        result = filesystem.read_file(str(binary_file))
        self.assertIsInstance(result, bytes)
        self.assertEqual(result, binary_content)



"""
Provides tools for getting the current date and time and waiting for a given number of seconds.
"""

import time
from typing import Annotated

# for docs
EMOJI = "🕒"


def current_time_and_date() -> str:
    """Get the current date and time."""

    return time.strftime("%H:%M%p %Z on %b %d, %Y")


def wait(
    seconds: Annotated[int, "The number of seconds to wait"],
) -> None:
    """Wait for a given number of seconds."""

    time.sleep(seconds)



import asyncio
import pathlib
import typing as t

import typer
from loguru import logger
from termcolor import colored

import nerve
from nerve.cli.defaults import (
    DEFAULT_AGENTS_LOAD_PATH,
)
from nerve.models import Configuration, Workflow

cli = typer.Typer(
    no_args_is_help=True,
    pretty_exceptions_enable=False,
    context_settings={"help_option_names": ["-h", "--help"]},
)


@cli.command(
    context_settings={"help_option_names": ["-h", "--help"]},
    help="List the agents available locally in $HOME/.nerve/agents or a custom path.",
)
def agents(
    path: t.Annotated[
        pathlib.Path,
        typer.Argument(help="Path to the agent or workflow to create"),
    ] = DEFAULT_AGENTS_LOAD_PATH,
) -> None:
    print(f"🧠 nerve v{nerve.__version__}")

    asyncio.run(show_agents(path))


async def show_agents(path: pathlib.Path) -> None:
    if not path.exists():
        logger.error(f"Path {path} does not exist")
        raise typer.Exit(1)

    if not path.is_dir():
        logger.error(f"Path {path} is not a directory")
        raise typer.Exit(1)

    print()
    print(f"📁 {path.absolute()}")

    anything = False
    for item in path.iterdir():
        if Workflow.is_workflow(item):
            print(f"   {item.name} " + colored("<workflow>", "blue"))
            anything = True
        elif Configuration.is_agent_config(item):
            print(f"   {item.name} " + colored("<agent>", "green"))
            anything = True

    if not anything:
        print(colored("no agents or workflows found", "light_grey"))



"""
Let the agent autonomously set the task as complete or failed.
"""
import typing as t

import nerve.runtime.state as state

# for docs
EMOJI = "✅"


def task_complete_success(
    reason: t.Annotated[
        str | None, "Optional reason why the task is complete or report of conclusive information."
    ] = None,
) -> None:
    """When your objective has been reached use this tool to set the task as complete."""

    state.set_task_complete(reason)


def task_failed(
    reason: t.Annotated[str, "The reason why the task is impossible"],
) -> None:
    """Use this tool if you determine that the given goal or task is impossible given the information you have."""

    state.set_task_failed(reason)



"""
Write primitives to the local filesystem.
"""

import os
from typing import Annotated

from nerve.tools.utils import path_acl

# for docs
EMOJI = "📂"

# if set, the agent will only have access to these paths
jail: list[str] = []


def create_file(
    path: Annotated[str, "The path to the file to create"],
    content: Annotated[
        str | None, "The content to write to the file, if not provided, the file will be created empty"
    ] = None,
) -> str:
    """Create a file on disk, if the file already exists, it will be overwritten."""

    path_acl(path, jail)

    response = ""

    # ensure parent directory exists
    parent_dir = os.path.dirname(path)
    if parent_dir and not os.path.exists(parent_dir):
        os.makedirs(parent_dir)
        response += f"Created parent directory {parent_dir}.\n"

    exists = os.path.exists(path)

    with open(path, "w") as f:
        f.write(content or "")

    if exists:
        response += f"File {path} updated.\n"
    else:
        response += f"File {path} created.\n"

    return response


def delete_file(path: Annotated[str, "The path to the file to delete"]) -> str:
    """Delete a file from disk."""

    path_acl(path, jail)

    os.remove(path)
    return f"File {path} deleted."



"""
Simulates the reasoning process at runtime.
"""

from typing import Annotated

import nerve.runtime.state as state

# for docs
EMOJI = "🧠"


def think(thought: Annotated[str, "A thought to think about"]) -> None:
    """
    Adhere strictly to this reasoning framework, ensuring thoroughness, precision, and logical rigor.

    ## Problem Decomposition

    Break the query into discrete, sequential steps.
    Explicitly state assumptions and context.

    ## Stepwise Analysis

    Address each step individually.
    Explain the rationale, principles, or rules applied (e.g., mathematical laws, linguistic conventions).
    Use examples, analogies, or intermediate calculations to illustrate reasoning.

    ## Validation & Error Checking

    Verify logical consistency at each step.
    Flag potential oversights, contradictions, or edge cases.
    Confirm numerical accuracy (e.g., recompute calculations).

    ## Synthesis & Conclusion

    Integrate validated steps into a coherent solution.
    Summarize key insights and ensure the conclusion directly addresses the original query.
    """
    state.append_to_knowledge("thoughts", thought)


def clear_thoughts() -> None:
    """If the reasoning process proved wrong, inconsistent or ineffective, clear your thoughts and start again."""
    state.clear_knowledge("thoughts")



"""
Let the agent interactively ask questions to the user in a structured way.
"""

from typing import Annotated

import inquirer  # type: ignore
from pydantic import Field

# for docs
EMOJI = "💬"


def ask_question(
    question: Annotated[
        str,
        Field(description="The question to ask the user.", examples=["What is your name?", "What is your age?"]),
    ],
) -> str:
    """Ask a question to the user."""

    print()
    return str(inquirer.prompt([inquirer.Text("question", message=question)])["question"]).strip()


def ask_for_confirmation(
    question: Annotated[
        str,
        Field(
            description="The question to ask the user.", examples=["Do you want to continue?", "Do you want to quit?"]
        ),
    ],
    default: Annotated[bool, Field(description="The default answer to the question.", examples=[True, False])] = False,
) -> str:
    """Ask a confirmation question to the user."""

    print()
    return (
        "YES" if inquirer.prompt([inquirer.Confirm("confirm", message=question, default=default)])["confirm"] else "NO"
    )


def ask_for_single_choice(
    question: Annotated[
        str,
        Field(
            description="The question to ask the user.",
            examples=["What is your favorite color?", "What is your favorite food?"],
        ),
    ],
    choices: Annotated[
        list[str],
        Field(description="The choices to offer the user.", examples=["red", "blue", "green"]),
    ],
) -> str:
    """Ask a single choice question to the user."""

    print()
    return str(inquirer.prompt([inquirer.List("choice", message=question, choices=choices)])["choice"]).strip()


def ask_for_multiple_choice(
    question: Annotated[
        str,
        Field(
            description="The question to ask the user.",
            examples=["What are your favorite colors?", "What are your favorite foods?"],
        ),
    ],
    choices: Annotated[
        list[str],
        Field(description="The choices to offer the user.", examples=["red", "blue", "green"]),
    ],
) -> str:
    """Ask a multiple choice question to the user."""

    print()
    return ", ".join(
        str(choice).strip()
        for choice in inquirer.prompt([inquirer.Checkbox("choices", message=question, choices=choices)])["choices"]
    )



import pathlib
import time

from loguru import logger

import nerve.runtime.state as state
from nerve.generation import WindowStrategy
from nerve.generation.conversation import FullHistoryStrategy
from nerve.models import Workflow
from nerve.runtime.agent import Agent
from nerve.runtime.shell import Shell

IS_ACTIVE: bool = False


class Flow:
    def __init__(
        self,
        actors: list[Agent],
        workflow: Workflow | None = None,
        max_steps: int = 500,
        max_cost: float = 10.0,
        timeout: int | None = None,
    ):
        global IS_ACTIVE

        if IS_ACTIVE:
            raise RuntimeError("A flow is already running")

        IS_ACTIVE = True

        # all actors in the flow
        self.actors = actors
        # workflow definition if set or None if we're running a single agent
        self.workflow = workflow
        # current active agent in the flow
        self.curr_actor_idx: int = 0
        self.curr_actor: Agent | None = None
        # current step from the beginning of the flow
        self.curr_step: int = 0
        # optional max steps to run
        self.max_steps: int = max_steps
        # max cost to run the flow
        self.max_cost: float = max_cost
        # optional timeout to run the flow
        self.timeout: int | None = timeout
        # start time of the flow
        self.started_at: float | None = None
        # interactive shell
        self.shell: Shell = Shell()

    @classmethod
    def build(
        cls,
        actors: list[Agent],
        max_steps: int = 500,
        max_cost: float = 10.0,
        timeout: int | None = None,
        start_state: dict[str, str] | None = None,
    ) -> "Flow":
        if start_state:
            state.update_variables(start_state)

        return cls(actors=actors, max_steps=max_steps, max_cost=max_cost, timeout=timeout)

    @classmethod
    def from_path(
        cls,
        input_path: pathlib.Path,
        window_strategy: WindowStrategy = FullHistoryStrategy(),
        max_steps: int = 500,
        max_cost: float = 10.0,
        timeout: int | None = None,
        start_state: dict[str, str] | None = None,
    ) -> "Flow":
        workflow = Workflow.from_path(input_path)
        root_path = input_path if input_path.is_dir() else input_path.parent

        actors = []
        for actor_name, actor in workflow.actors.items():
            # determine actor task file
            task_file_path = (root_path / actor_name).with_suffix(".yml")
            if not task_file_path.exists():
                task_file_path = root_path / actor_name

            actors.append(Agent.create_from_file(actor.generator, task_file_path, window_strategy))

        if start_state:
            state.update_variables(start_state)

        return cls(
            actors=actors,
            workflow=workflow,
            max_steps=max_steps,
            max_cost=max_cost,
            timeout=timeout,
        )

    async def _setup_if_needed(self) -> None:
        if self.started_at is None:
            self.started_at = time.time()

        if self.curr_actor is None:
            self.curr_actor = self.actors[self.curr_actor_idx]

            state.set_tools({tool.__name__: tool for tool in self.curr_actor.runtime.tools})
            state.set_defaults(self.curr_actor.configuration.defaults)
            state.on_task_started(self.curr_actor)

    async def step(self) -> None:
        await self._setup_if_needed()

        if self.done():
            state.on_event("flow_complete", {"steps": self.curr_step, "usage": state.get_usage()})
            return

        state.on_event("step_started", {"step": self.curr_step, "usage": state.get_usage()})

        step_usage = await self.curr_actor.step()  # type: ignore
        logger.debug(f"step usage: {step_usage}")

        # increment total usage
        state.update_usage(step_usage)
        state.on_event("step_complete", {"step": self.curr_step, "usage": state.get_usage()})

        if state.is_active_task_done():
            logger.debug(f"task {self.curr_actor.runtime.name} complete")  # type: ignore
            self.curr_actor_idx += 1
            self.curr_actor = None
            state.reset()

        self.curr_step += 1

    def done(self) -> bool:
        if self.curr_actor_idx >= len(self.actors):
            return True

        if self.max_steps is not None and self.curr_step > self.max_steps:
            state.on_max_steps_reached()
            return True

        usage = state.get_usage()
        if self.max_cost is not None and usage.cost is not None and usage.cost > self.max_cost:
            state.on_max_cost_reached()
            return True

        if self.timeout is not None and self.started_at is not None and time.time() - self.started_at > self.timeout:
            state.on_timeout()
            return True

        return False

    async def run(self) -> None:
        state.on_event(
            "flow_started",
            {
                "flow": self,
                "state": state.as_dict(),
            },
        )

        while not self.done():
            await self._setup_if_needed()
            if self.curr_actor:
                await self.shell.interact_if_needed(self.curr_actor)
            await self.step()

        state.on_event(
            "flow_complete",
            {
                "workflow": self.workflow,
                "steps": self.curr_step - 1,
                "usage": state.get_usage(),
                "state": state.as_dict(),
            },
        )



import pathlib
import tempfile
import unittest

from nerve.models import Configuration


class TestConfiguration(unittest.TestCase):
    def test_is_legacy_with_system_prompt(self) -> None:
        """Test that is_legacy returns True when system_prompt is set"""
        config = Configuration(system_prompt="This is a legacy prompt")
        self.assertTrue(config.is_legacy)

    def test_is_legacy_without_system_prompt(self) -> None:
        """Test that is_legacy returns False when system_prompt is None"""
        config = Configuration(agent="Agent prompt", task="Task description")
        self.assertFalse(config.is_legacy)

    def test_is_legacy_with_empty_config(self) -> None:
        """Test that is_legacy returns False with default configuration"""
        config = Configuration()
        self.assertFalse(config.is_legacy)

    def test_system_prompt_excluded_from_serialization(self) -> None:
        """Test that system_prompt is excluded when serializing the configuration"""
        config = Configuration(system_prompt="This should be excluded", agent="Agent prompt")
        serialized = config.model_dump()
        self.assertNotIn("system_prompt", serialized)

    def test_system_prompt_loaded_from_file(self) -> None:
        """Test that system_prompt is loaded if present in a file"""

        with tempfile.NamedTemporaryFile(suffix=".yml", delete=False) as temp_file:
            temp_file.write(b"system_prompt: Legacy prompt")
            temp_path = pathlib.Path(temp_file.name)

        try:
            loaded_config = Configuration.from_path(temp_path)

            self.assertEqual(loaded_config.system_prompt, "Legacy prompt")
            self.assertTrue(loaded_config.is_legacy)
        finally:
            temp_path.unlink()



from pathlib import Path


def is_path_allowed(path_to_check: str, jail: list[str] | None = None) -> bool:
    if not jail:
        return True

    # https://stackoverflow.com/questions/3812849/how-to-check-whether-a-directory-is-a-sub-directory-of-another-directory
    path = Path(path_to_check).resolve().absolute()
    for allowed_path in jail:
        allowed = Path(allowed_path).resolve().absolute()
        if path == allowed or allowed in path.parents:
            return True

    return False


def path_acl(path_to_check: str, jail: list[str] | None = None) -> None:
    if not is_path_allowed(path_to_check, jail):
        raise ValueError(f"access to path {path_to_check} is not allowed, only allowed paths are: {jail}")


def maybe_text(output: bytes) -> str | bytes:
    try:
        return output.decode("utf-8").strip()
    except UnicodeDecodeError:
        return output



import asyncio
import pathlib
import pkgutil
import typing as t

import inquirer  # type: ignore
import typer
from pydantic_yaml import to_yaml_str

import nerve
from nerve.cli.defaults import (
    DEFAULT_AGENT_PATH,
    DEFAULT_AGENT_SYSTEM_PROMPT,
    DEFAULT_AGENT_TASK,
    DEFAULT_AGENT_TOOLS,
    DEFAULT_PROMPTS_LOAD_PATH,
)
from nerve.models import Configuration, Tool

cli = typer.Typer(
    no_args_is_help=True,
    pretty_exceptions_enable=False,
    context_settings={"help_option_names": ["-h", "--help"]},
)


@cli.command(
    context_settings={"help_option_names": ["-h", "--help"]},
    help="Guided procedure for creating a new agent.",
)
def create(
    path: t.Annotated[
        pathlib.Path,
        typer.Argument(help="Path to the agent or workflow to create"),
    ] = DEFAULT_AGENT_PATH,
    task: t.Annotated[
        str | None,
        typer.Option("--task", "-t", help="Task to create the agent for"),
    ] = None,
    default: t.Annotated[
        bool,
        typer.Option("--default", "-d", help="Use default values."),
    ] = False,
) -> None:
    print(f"🧠 nerve v{nerve.__version__}")

    asyncio.run(create_agent(path.absolute(), task=task, default=default))


def _get_available_namespaces(defaults: list[str]) -> tuple[list[str], list[str]]:
    import nerve.tools.namespaces as namespaces

    available_namespaces = []
    default_entries = []

    for _, modname, _ in pkgutil.iter_modules(namespaces.__path__):
        if modname[0] != "_" and "test" not in modname:
            try:
                module = __import__(f"nerve.tools.namespaces.{modname}", fromlist=[""])
                doc = module.__doc__ or ""
                entry = f"{modname} - {doc.strip()}"
                available_namespaces.append(entry)
                if modname in defaults:
                    default_entries.append(entry)
            except Exception:
                pass

    return available_namespaces, default_entries


def _resolve_system_prompt(system_prompt: str) -> str:
    if system_prompt.startswith("@"):
        system_prompt_file = DEFAULT_PROMPTS_LOAD_PATH / (system_prompt[1:] + ".md")
        if system_prompt_file.exists():
            print(f"🔍 loading system prompt from {system_prompt_file}")
            with open(system_prompt_file) as f:
                system_prompt = f.read()
        else:
            system_prompt_nested = DEFAULT_PROMPTS_LOAD_PATH / system_prompt[1:] / "system.md"
            if system_prompt_nested.exists():
                print(f"🔍 loading system prompt from {system_prompt_nested}")
                with open(system_prompt_nested) as f:
                    system_prompt = f.read()

    return system_prompt.strip()


def _collect_user_prompts() -> list[str]:
    prompts: list[str] = []

    if not DEFAULT_PROMPTS_LOAD_PATH.exists():
        return prompts

    for item in DEFAULT_PROMPTS_LOAD_PATH.iterdir():
        if item.is_dir():
            system_file = item.joinpath("system.md")
            if system_file.exists():
                prompts.append(f"@{item.name}")
        elif item.is_file() and item.suffix == ".md" and item.name != "system.md":
            prompts.append(f"@{item.stem}")

    return sorted(prompts)


# TODO: create a doc page for this.


async def create_agent(path: pathlib.Path, task: str | None = None, default: bool = False) -> None:
    if path.exists():
        print(f"❌ {path} already exists.")
        exit(1)

    available_namespaces, defaults = _get_available_namespaces(DEFAULT_AGENT_TOOLS)
    user_prompts = _collect_user_prompts()

    if default:
        answers = {
            "path": path,
            "system_prompt": DEFAULT_AGENT_SYSTEM_PROMPT,
            "prompt": task or DEFAULT_AGENT_TASK.replace("{{", "{").replace("}}", "}"),
            "tools": DEFAULT_AGENT_TOOLS,
        }
    else:
        print()
        basic_system_prompt_prompt = inquirer.Text(
            "system_prompt", message="System prompt", default=DEFAULT_AGENT_SYSTEM_PROMPT
        )
        questions = [
            inquirer.Path("path", message="Path", default=str(path)),
            basic_system_prompt_prompt
            if not user_prompts
            else inquirer.List(
                "system_prompt",
                message="System prompt",
                choices=["use custom, or:"] + user_prompts,
            ),
            inquirer.Text("prompt", message="Task", default=task or DEFAULT_AGENT_TASK),
            inquirer.Checkbox(
                "tools",
                message="Built-in tools",
                choices=available_namespaces,
                carousel=True,
                default=defaults,
            ),
        ]

        answers = inquirer.prompt(questions)
        answers["tools"] = [tool.split(" - ")[0] for tool in answers["tools"]]  # type: ignore

        if user_prompts:
            answer = str(answers["system_prompt"])
            if answer == "use custom, or:":
                a = inquirer.prompt([basic_system_prompt_prompt])
                answers["system_prompt"] = _resolve_system_prompt(str(a["system_prompt"]))
            else:
                answers["system_prompt"] = _resolve_system_prompt(answer)
        else:
            answers["system_prompt"] = _resolve_system_prompt(str(answers["system_prompt"]))

    example_tool = Tool(
        name="get_weather",
        description="Get the current weather in a given place.",
        arguments=[Tool.Argument(name="place", description="The place to get the weather of.", example="Rome")],
        tool="curl wttr.in/{{ place }}",
    )

    config = Configuration(
        agent=answers["system_prompt"],  # type: ignore
        task=answers["prompt"],  # type: ignore
        using=answers["tools"],  # type: ignore
    )

    config_with_tool = Configuration(
        tools=[example_tool],
    )

    example_tool_lines = ["# You can add custom tools like this:", "#"] + [
        f"# {line}"
        for line in to_yaml_str(config_with_tool, exclude_defaults=True, default_flow_style=False).split("\n")
        if line.strip() != ""
    ]

    agent_code = (
        to_yaml_str(config, exclude_defaults=True, default_flow_style=False) + "\n" + "\n".join(example_tool_lines)
    )

    path = pathlib.Path(str(answers["path"]))
    if path.suffix == "":
        if not path.exists():
            path.mkdir(parents=True)
        path = path / "agent.yml"

    with open(path, "w") as f:
        f.write(agent_code)

    print(f"🤖 agent saved to {path}")

    print()
    answers = inquirer.prompt([inquirer.Confirm("start", message="Start the agent now?", default=True)])
    if answers["start"]:
        import os

        os.system(f"nerve run {path}")



from unittest.mock import MagicMock

import pytest

from nerve.runtime.flow import Flow


def test_flow_singleton() -> None:
    # Create a mock agent
    mock_agent = MagicMock()

    # First flow creation should succeed
    flow1 = Flow(actors=[mock_agent], max_steps=10)
    assert flow1 is not None

    # Second flow creation should raise RuntimeError
    with pytest.raises(RuntimeError) as excinfo:
        flow2 = Flow(actors=[mock_agent], max_steps=10)  # noqa: F841

    # Verify the error message
    assert "A flow is already running" in str(excinfo.value)



import typing as t
import uuid

import litellm
from loguru import logger

from nerve.generation import Engine, WindowStrategy
from nerve.generation.conversation import SlidingWindowStrategy
from nerve.models import Usage
from nerve.runtime import state


class LiteLLMEngine(Engine):
    def __init__(
        self,
        generator_id: str,
        window_strategy: WindowStrategy,
        tools: list[t.Callable[..., t.Any]] | None = None,
    ):
        super().__init__(generator_id, window_strategy, tools)

        # until this is not fixed, ollama needs special treatment: https://github.com/BerriAI/litellm/issues/6353
        self.is_ollama = "ollama" in self.generator_id
        self.reduced_window_size = 25

        if not self.is_ollama:
            if self.tools and not litellm.supports_function_calling(model=self.generator_id):  # type: ignore
                logger.error(f"model {self.generator_id} does not support function calling")
                exit(1)

        else:
            import ollama

            self.ollama_model = "/".join(self.generator_id.split("/")[1:])
            self.ollama_client = ollama.AsyncClient(host=self.api_base)

            logger.debug(f"using ollama client for model {self.ollama_model}")

    async def _ollama_generate(
        self, conversation: list[dict[str, t.Any]], tools_schema: list[dict[str, t.Any]] | None
    ) -> tuple[Usage, t.Any]:
        response = await self.ollama_client.chat(
            model=self.ollama_model,
            messages=conversation,
            tools=tools_schema,
            **self.generator_params,
        )
        return Usage(
            prompt_tokens=0,
            completion_tokens=0,
            total_tokens=0,
        ), response.message

    async def _litellm_generate(
        self, conversation: list[dict[str, t.Any]], tools_schema: list[dict[str, t.Any]] | None
    ) -> tuple[Usage, t.Any]:
        try:
            # litellm.set_verbose = True
            response = litellm.completion(
                model=self.generator_id,
                messages=conversation,
                tools=tools_schema,
                tool_choice="auto" if tools_schema else None,
                api_base=self.api_base,
                **self.generator_params,
            )

            return Usage(
                prompt_tokens=response.usage.prompt_tokens,
                completion_tokens=response.usage.completion_tokens,
                total_tokens=response.usage.total_tokens,
                cost=response._hidden_params.get("response_cost", None),
            ), response.choices[0].message
        except litellm.AuthenticationError as e:  # type: ignore
            logger.error(e)
            exit(1)

    async def _generate(
        self, conversation: list[dict[str, t.Any]], tools_schema: list[dict[str, t.Any]] | None
    ) -> tuple[Usage, t.Any]:
        if self.is_ollama:
            # https://github.com/BerriAI/litellm/issues/6353
            return await self._ollama_generate(conversation, tools_schema)
        else:
            return await self._litellm_generate(conversation, tools_schema)

    async def _get_conversation(
        self, system_prompt: str | None, user_prompt: str, extra_message: str | None
    ) -> list[dict[str, t.Any]]:
        # @ system prompt and user prompt always included
        conversation = [{"role": "system", "content": system_prompt}] if system_prompt else []
        conversation.append({"role": "user", "content": user_prompt})
        conversation.extend(await self.window_strategy.get_window(self.history))

        if extra_message:
            conversation.append({"role": "user", "content": extra_message})

        logger.debug(f"{self.window_strategy} | conv size: {len(conversation)}")

        return conversation

    async def _generate_next_message(
        self,
        system_prompt: str | None,
        user_prompt: str,
        extra_tools: dict[str, t.Callable[..., t.Any]] | None = None,
        extra_message: str | None = None,
    ) -> tuple[Usage, t.Any]:
        # build chat history
        conversation = await self._get_conversation(system_prompt, user_prompt, extra_message)
        # build json schema for available tools
        extra_tools = extra_tools or {}
        tools_schema = self._get_extended_tooling_schema(extra_tools)

        try:
            # get next message
            return await self._generate(conversation, tools_schema)
        except litellm.ContextWindowExceededError as e:  # type: ignore
            logger.debug(e)
            if self.reduced_window_size > 0:
                self.reduced_window_size -= 1
                self.window_strategy = SlidingWindowStrategy(self.reduced_window_size)
                logger.error(f"context window exceeded, adjusting window strategy: {self.window_strategy}")
                return await self._generate_next_message(system_prompt, user_prompt, extra_tools)

            else:
                logger.error("context window exceeded")
                return Usage(
                    prompt_tokens=0,
                    completion_tokens=0,
                    total_tokens=0,
                ), None
        except Exception as e:
            logger.error(e)
            return Usage(
                prompt_tokens=0,
                completion_tokens=0,
                total_tokens=0,
            ), None

    async def step(
        self,
        system_prompt: str | None,
        user_prompt: str,
        extra_tools: dict[str, t.Callable[..., t.Any]] | None = None,
        extra_message: str | None = None,
    ) -> Usage:
        extra_tools = extra_tools or {}

        # get next message
        usage, message = await self._generate_next_message(system_prompt, user_prompt, extra_tools, extra_message)
        if message is None:
            return usage

        # collect responses
        has_tools = len(extra_tools) > 0 or len(self.tools) > 0
        responses: list[dict[str, t.Any]] = []

        if has_tools and not message.tool_calls:
            # no tool calls, just return the text response
            responses = [self._get_text_response(str(message.content))]

        elif message.tool_calls:
            logger.debug(message.tool_calls)
            # for each tool call
            for tool_call in message.tool_calls:
                # resolve and execute the tool call
                responses.extend(
                    await self._process_tool_call(
                        tool_call.id if hasattr(tool_call, "id") else str(uuid.uuid4()),
                        tool_call.function.name or "",
                        tool_call.function.arguments,
                        extra_tools,
                    )
                )

                # break early from multiple tool calls if the task is complete
                if state.is_active_task_done():
                    logger.debug(f"task {self.generator_id} complete")
                    break

        # add tool call + per-call response messages
        self.history.append(message.__dict__)
        self.history.extend(responses)

        return usage



import pathlib
import typing as t

from loguru import logger

from nerve.models import Tool
from nerve.tools import compiler


class Runtime:
    def __init__(self, name: str, generator: str, working_dir: pathlib.Path):
        # task unique identifier
        self.name = name
        # which model will handle this task
        self.generator = generator
        # the working directory for the task
        self.working_dir = working_dir
        # full chat history of the task
        self.history: list[t.Any] = []
        # available tools
        self.tools: list[t.Callable[..., t.Any]] = []

    @classmethod
    def build(
        cls,
        working_dir: pathlib.Path,
        name: str,
        generator: str,
        using: list[str],
        jail: dict[str, list[str]],
        tools: list[Tool | t.Callable[..., t.Any]],
    ) -> "Runtime":
        runtime = cls(name=name, generator=generator, working_dir=working_dir)

        # import tools from builtin namespaces
        ns_tools = compiler.get_tools_from_namespaces(using, jail)
        if ns_tools:
            logger.debug(f"🧰 importing {len(ns_tools)} tools from: {using}")
            runtime.tools.extend(ns_tools)

        # import custom tools from yaml definition
        yml_tools = compiler.get_tools_from_yml(
            working_dir,
            [tool for tool in tools if isinstance(tool, Tool) and not tool.path],
        )
        if yml_tools:
            logger.debug(f"🧰 importing {len(yml_tools)} tools from: {working_dir}")
            runtime.tools.extend(yml_tools)

        # import custom tools from files
        py_tools = compiler.get_tools_from_files(
            working_dir,
            [tool.path for tool in tools if isinstance(tool, Tool) and tool.path],
        )
        if py_tools:
            logger.debug(f"🧰 importing {len(py_tools)} tools from: {working_dir}")
            runtime.tools.extend(py_tools)

        # import custom tools from functions (when used as sdk)
        funcs = [compiler.wrap_tool_function(tool) for tool in tools if callable(tool)]
        if funcs:
            logger.debug(f"🧰 importing {len(funcs)} custom tools from functions")
            runtime.tools.extend(funcs)

        return runtime



import typer

import nerve
from nerve.cli.agents import cli as agents_cli
from nerve.cli.create import cli as create_cli
from nerve.cli.namespaces import cli as namespaces_cli
from nerve.cli.replay import cli as replay_cli
from nerve.cli.run import cli as run_cli

cli = typer.Typer(
    no_args_is_help=True,
    pretty_exceptions_enable=False,
    context_settings={"help_option_names": ["-h", "--help"]},
)

cli.add_typer(agents_cli)
cli.add_typer(create_cli)
cli.add_typer(run_cli)
cli.add_typer(replay_cli)
cli.add_typer(namespaces_cli)

# TODO Workflows 2.0: shared events bus (IPC and network) for multiple agents (one input variable = one event).
# Tools:
# advertise: ask the agent to advertise itself on the events bus
# request: allow the agent to request something from another agent
# respond: allow the agent to respond to a request
# at every step the agent goes requests = bus.query()
# TODO: browser-use: each interactable html generates tools at runtime.


@cli.command(
    context_settings={"help_option_names": ["-h", "--help"]},
    help="Print the version and exit.",
)
def version() -> None:
    import platform
    import sys

    print(f"platform: {platform.system().lower()} ({platform.machine()})")
    print(f"python:   {sys.version_info.major}.{sys.version_info.minor}.{sys.version_info.micro}")
    print(f"nerve:    {nerve.__version__}")



import os
import tempfile
import unittest
from pathlib import Path

from nerve.tools.utils import is_path_allowed, maybe_text, path_acl


class TestUtils(unittest.TestCase):
    def test_maybe_text_with_valid_utf8(self) -> None:
        # Test with valid UTF-8 string
        input_bytes = b"Hello, world!"
        result = maybe_text(input_bytes)
        self.assertEqual(result, "Hello, world!")
        self.assertIsInstance(result, str)

    def test_maybe_text_with_invalid_utf8(self) -> None:
        # Test with invalid UTF-8 sequence
        input_bytes = bytes([0xFF, 0xFE, 0xFD])  # Invalid UTF-8 sequence
        result = maybe_text(input_bytes)
        self.assertEqual(result, input_bytes)
        self.assertIsInstance(result, bytes)

    def test_maybe_text_with_empty_bytes(self) -> None:
        # Test with empty bytes
        input_bytes = b""
        result = maybe_text(input_bytes)
        self.assertEqual(result, "")
        self.assertIsInstance(result, str)

    def test_maybe_text_with_mixed_content(self) -> None:
        # Test with mixed content (valid UTF-8 + invalid sequence)
        input_bytes = b"Valid text" + bytes([0xFF, 0xFE]) + b"more text"
        result = maybe_text(input_bytes)
        self.assertEqual(result, input_bytes)
        self.assertIsInstance(result, bytes)

    def test_maybe_text_with_non_bytes_input(self) -> None:
        # Test with non-bytes input (should handle gracefully or raise appropriate error)
        with self.assertRaises(AttributeError):
            maybe_text("already a string")  # type: ignore

    def test_path_acl_with_no_jail(self) -> None:
        # Test with empty jail list
        jail: list[str] = []
        path = "/some/random/path"
        self.assertTrue(is_path_allowed(path, jail))

        # Test path_acl with empty jail
        path_acl(path, jail)  # Should not raise an exception with empty jail

    def test_path_acl_with_jail(self) -> None:
        # Test with jail containing paths
        jail = ["/allowed/path", "/another/allowed"]

        # Test allowed path
        allowed_path = "/allowed/path/file.txt"
        self.assertTrue(is_path_allowed(allowed_path, jail))
        path_acl(allowed_path, jail)  # Should not raise an exception

        # Test another allowed path
        another_allowed = "/another/allowed/subdir/file.txt"
        self.assertTrue(is_path_allowed(another_allowed, jail))
        path_acl(another_allowed, jail)  # Should not raise an exception

        # Test disallowed path
        disallowed_path = "/disallowed/path/file.txt"
        self.assertFalse(is_path_allowed(disallowed_path, jail))

        # Test path_acl with disallowed path (should raise ValueError)
        with self.assertRaises(ValueError):
            path_acl(disallowed_path, jail)

    def test_path_acl_with_symlinks(self) -> None:
        # Create temporary directories for testing
        with tempfile.TemporaryDirectory() as allowed_dir, tempfile.TemporaryDirectory() as disallowed_dir:
            allowed_path = Path(allowed_dir)
            disallowed_path = Path(disallowed_dir)

            # Create a symlink inside allowed directory pointing to disallowed directory
            symlink_path = allowed_path / "symlink_to_outside"
            os.symlink(str(disallowed_path), str(symlink_path))

            jail = [str(allowed_path)]

            # Test direct access to allowed path
            self.assertTrue(is_path_allowed(str(allowed_path), jail))

            # Test symlink that resolves outside jail
            symlink_file_path = str(symlink_path / "some_file.txt")
            self.assertFalse(is_path_allowed(symlink_file_path, jail))

            with self.assertRaises(ValueError):
                path_acl(symlink_file_path, jail)



"""
Computer use primitives for mouse, keyboard, and screen.
"""

import base64
import io
import typing as t

import pyautogui as px
import pyperclip

# this is an extra feature, so we need to indicate it
OPTIONAL_FEATURE = "computer_use"
# for docs
EMOJI = "💻"

# max screenshot width
MAX_WIDTH = 1280
# typing delay in ms
TYPING_DELAY_MS = 12

_width = 0
_height = 0
_display_num = 0
_scale_factor = 1.0
_target_width = 0
_target_height = 0
_scaling_enabled = False


def _init_scaling() -> None:
    global _width, _height, _display_num, _scale_factor, _target_width, _target_height

    if _width is None:
        screen_size = px.size()
        _width = int(screen_size[0])
        _height = int(screen_size[1])
        _display_num = None  # Not used on MacOS

        if _scaling_enabled and _width > MAX_WIDTH:
            _scale_factor = MAX_WIDTH / _width
            _target_width = MAX_WIDTH
            _target_height = int(_height * _scale_factor)
        else:
            _scale_factor = 1.0
            _target_width = _width
            _target_height = _height


def _scale_coordinates(x: int, y: int) -> tuple[int, int]:
    if not _scaling_enabled:
        return x, y

    x_scaling_factor = _width / _target_width
    y_scaling_factor = _height / _target_height

    return round(x / x_scaling_factor), round(y / y_scaling_factor)


async def screenshot() -> dict[str, t.Any]:
    """Take a screenshot of the current screen."""

    _init_scaling()

    screenshot = px.screenshot()
    if _scaling_enabled and _scale_factor < 1.0:
        screenshot = screenshot.resize((_target_width, _target_height))

    img_buffer = io.BytesIO()

    screenshot.save(img_buffer, format="PNG", optimize=True)
    img_buffer.seek(0)
    base64_image = base64.b64encode(img_buffer.read()).decode()

    return {
        "type": "image_url",
        "image_url": {"url": f"data:image/png;base64,{base64_image}"},
    }


async def get_cursor_position() -> str:
    """Get the current mouse position."""

    x, y = px.position()
    x, y = _scale_coordinates(int(x), int(y))

    return f"({x}, {y})"


async def mouse_move(
    x: t.Annotated[int, "The x coordinate to move to"],
    y: t.Annotated[int, "The y coordinate to move to"],
) -> None:
    """Move the mouse to the given coordinates."""

    px.moveTo(x, y)


async def mouse_left_click() -> None:
    """Click the left mouse button at the current mouse position."""

    px.click(button="left")


async def mouse_left_click_drag(
    x: t.Annotated[int, "The x coordinate to move to"],
    y: t.Annotated[int, "The y coordinate to move to"],
) -> None:
    """Click and drag the left mouse button from the current mouse position to the given coordinates."""

    px.mouseDown(button="left")
    px.moveTo(x, y)
    px.mouseUp(button="left")


async def mouse_right_click() -> None:
    """Click the right mouse button at the current mouse position."""

    px.click(button="right")


async def mouse_middle_click() -> None:
    """Click the middle mouse button at the current mouse position."""

    px.click(button="middle")


async def mouse_double_click() -> None:
    """Double click the left mouse button at the current mouse position."""

    px.doubleClick()


async def mouse_scroll(
    x: t.Annotated[int, "The x coordinate to move to"],
    y: t.Annotated[int, "The y coordinate to move to"],
) -> None:
    """Scroll the mouse wheel in the given direction."""

    px.scroll(x, y)


async def keyboard_press_hotkeys(
    keys: t.Annotated[str, "The hotkey sequence to press (like 'ctrl+shift+cmd+space')"],
) -> None:
    """Press one or more hotkeys on the keyboard."""

    key_sequence = list(map(str.strip, keys.lower().split("+")))
    replace = {
        "control": "ctrl",
        "option": "alt",
        "super": "command",
        "cmd": "command",
        "return": "enter",
        "escape": "esc",
        "spacebar": "space",
    }
    key_sequence = [replace.get(key, key) for key in key_sequence]

    px.hotkey(*key_sequence)


async def keyboard_type(text: t.Annotated[str, "The text to type"]) -> None:
    """Type the given text on the keyboard."""

    # px.typewrite(text, interval=TYPING_DELAY_MS / 1000.0)

    # https://github.com/asweigart/pyautogui/issues/259
    pyperclip.copy(text)
    await keyboard_press_hotkeys("cmd+v")
    pyperclip.copy("")



import pkgutil

import typer
from termcolor import colored

import nerve
import nerve.tools.namespaces as ns_pkg

cli = typer.Typer(
    no_args_is_help=True,
    pretty_exceptions_enable=False,
    context_settings={"help_option_names": ["-h", "--help"]},
)


@cli.command(
    context_settings={"help_option_names": ["-h", "--help"]},
    help="List all the available tools namespaces.",
)
def namespaces() -> None:
    print(f"🧠 nerve v{nerve.__version__}")
    print()

    for _, modname, _ in pkgutil.iter_modules(ns_pkg.__path__):
        if modname[0] != "_" and "test" not in modname:
            module = __import__(f"nerve.tools.namespaces.{modname}", fromlist=[""])

            doc = module.__doc__ or ""
            print(f"{module.EMOJI} {colored(modname, 'magenta')} - {colored(doc.strip(), 'dark_grey')}")



"""
Let the agent create its own tools in Python.
"""

import typing as t
from typing import Annotated

from loguru import logger
from pydantic import Field

from nerve.runtime import state
from nerve.tools.compiler import wrap_tool_function

# for docs
EMOJI = "🔧"


def create_tool(
    code: Annotated[
        str,
        Field(
            description="The Python code to create the tool.",
            examples=[
                '''
# import any standard library you need
from typing import Annotated

def this_will_be_the_tool_name(
    argument_1: Annotated[str, "The first argument"],
    argument_2: Annotated[str, "The second argument"],
) -> str:
    """ALWAYS add a docstring to the tool."""
    return "This is the return value of the tool."
'''
            ],
        ),
    ],
) -> None:
    """Create a new tool or redefine an existing one by defining it as an annotated Python function.
    Use this tool to implement the missing functionalities you need to perform your task."""

    func_namespace: dict[str, t.Any] = {}
    exec(code, func_namespace)

    for name, value in func_namespace.items():
        if name[0] == "_":
            continue

        elif not callable(value):
            continue

        elif value.__module__ is not None:
            continue

        logger.debug(f"creating tool: {name}")
        tool_fn = wrap_tool_function(value)
        state.set_extra_tool(tool_fn)



import asyncio
import base64
import functools
import importlib
import inspect
import os
import pathlib
import time
import typing as t

import jinja2
from loguru import logger
from termcolor import colored

from nerve.models import Tool
from nerve.runtime import state


def wrap_tool_function(func: t.Callable[..., t.Any], mime: str | None = None) -> t.Callable[..., t.Any]:
    """
    Creates a wrapper around a function that logs the function call and its result.
    """

    async def wrapper(*args: t.Any, **kwargs: t.Any) -> t.Any:
        logger.debug(f"calling {func.__name__} ...")

        state.on_before_tool_called(func.__name__, kwargs)

        started_at = time.time()
        error = None
        try:
            result = func(*args, **kwargs)
            # check if the tool function returned a coroutine
            if asyncio.iscoroutine(result):
                result = await result

        except Exception as e:
            # import traceback
            # error_trace = traceback.format_exc()
            # logger.error(f"{func.__name__}: {e}")
            # logger.error(f"{error_trace}")
            result = f"ERROR in {func.__name__}: {e}"
            error = str(e)
            logger.error(colored(f"{func.__name__}: {e}", "red"))

        finished_at = time.time()

        state.on_tool_called(started_at, finished_at, func.__name__, kwargs, result, error)

        if mime:
            if mime.startswith("image/"):
                result = {
                    "type": "image_url",
                    "image_url": {"url": f"data:{mime};base64,{base64.b64encode(result).decode('utf-8')}"},
                }
            else:
                logger.error(f"tool {func.__name__} references an unsupported mime type: {mime}")
                exit(1)

        return result

    # Preserve the function's metadata
    wrapper = functools.wraps(func)(wrapper)

    return wrapper


def get_tools_from_namespace(namespace: str, jail: list[str]) -> list[t.Callable[..., t.Any]]:
    try:
        importlib.util.find_spec(f"nerve.tools.namespaces.{namespace}")
    except ImportError as err:
        raise ImportError(f"namespace {namespace} not found") from err

    try:
        module = __import__(f"nerve.tools.namespaces.{namespace}", fromlist=[""])
        if jail:
            for jailed_path in jail:
                jailed_path = state.interpolate(jailed_path)
                jailed_path = os.path.abspath(jailed_path)
                module.jail.append(jailed_path)
                logger.debug(f"namespace {namespace} jailed to: {jailed_path}")

        module_tools = [
            wrap_tool_function(func)
            for (name, func) in inspect.getmembers(module, inspect.isfunction)
            if name[0] != "_" and func.__module__ == module.__name__
        ]
        logger.debug(f"importing {len(module_tools)} tools from {namespace} namespace")
        return module_tools
    except ImportError as err:
        raise ImportError(f"could not import {namespace}: {err}") from err


def get_tools_from_namespaces(
    namespaces: list[str],
    jail: dict[str, list[str]],
) -> list[t.Callable[..., t.Any]]:
    tools = []

    for namespace in namespaces:
        tools.extend(get_tools_from_namespace(namespace, jail.get(namespace, [])))

    return tools


def get_tools_from_file(working_dir: pathlib.Path, file: str) -> list[t.Callable[..., t.Any]]:
    tool_path = pathlib.Path(file)
    if not tool_path.is_absolute():
        tool_path = working_dir / tool_path

    if not tool_path.exists():
        tool_path = tool_path.with_suffix(".py")
        if not tool_path.exists():
            raise FileNotFoundError(f"tool {file} does not exist")

    spec = importlib.util.spec_from_file_location(tool_path.stem, tool_path)
    module = importlib.util.module_from_spec(spec)  # type: ignore
    spec.loader.exec_module(module)  # type: ignore

    module_tools = [
        wrap_tool_function(func)
        for (name, func) in inspect.getmembers(module, inspect.isfunction)
        if name[0] != "_" and func.__module__ == module.__name__
    ]

    if module_tools:
        logger.debug(f"importing {len(module_tools)} tools from {tool_path}")

    return module_tools


def get_tools_from_files(working_dir: pathlib.Path, files: list[str]) -> list[t.Callable[..., t.Any]]:
    tools = []

    for file in files:
        tools.extend(get_tools_from_file(working_dir, file))

    return tools


def get_tool_from_yml(working_dir: pathlib.Path, tool: Tool) -> t.Callable[..., t.Any]:
    # load the template from the same directory as this script
    template_path = os.path.join(os.path.dirname(__file__), "body.j2")
    with open(template_path) as f:
        template_content = f.read()

    func_body = (
        jinja2.Environment().from_string(template_content).render(tool=tool, working_dir=str(working_dir.absolute()))
    )

    logger.debug(f"compiling tool {tool.name} as:\n{func_body}")

    func_namespace: dict[str, t.Any] = {}
    exec(func_body, func_namespace)

    return wrap_tool_function(func_namespace[tool.name], tool.mime)


def get_tools_from_yml(working_dir: pathlib.Path, yml_tools: list[Tool]) -> list[t.Callable[..., t.Any]]:
    tools = []

    for tool in yml_tools:
        tools.append(get_tool_from_yml(working_dir, tool))

    if tools:
        logger.debug(f"importing {len(tools)} tools from {working_dir}")

    return tools



import asyncio
import json
import pathlib
import typing as t

import typer
from loguru import logger

import nerve
from nerve.runtime import logging
from nerve.runtime.events import Event

cli = typer.Typer(
    no_args_is_help=True,
    pretty_exceptions_enable=False,
    context_settings={"help_option_names": ["-h", "--help"]},
)


@cli.command(
    context_settings={"help_option_names": ["-h", "--help"]},
    no_args_is_help=True,
    help="Replay a trace file.",
)
def play(
    trace_path: t.Annotated[
        pathlib.Path,
        typer.Argument(help="Trace file to replay"),
    ] = pathlib.Path("trace.jsonl"),
    fast: t.Annotated[
        bool,
        typer.Option("--fast", "-f", help="Do not sleep between events"),
    ] = False,
) -> None:
    logging.init(level="INFO")
    logger.info(f"🧠 nerve v{nerve.__version__}")

    asyncio.run(replay(trace_path, fast))


async def replay(
    trace_path: pathlib.Path,
    fast: bool,
) -> None:
    logger.info(f"▶️  replaying {trace_path} ...")

    with open(trace_path) as f:
        prev: Event | None = None

        for line in f:
            event = Event(**json.loads(line))

            if prev and not fast:
                delay = event.timestamp - prev.timestamp
                await asyncio.sleep(delay)

            logging.log_event_to_terminal(event)

            prev = event



import base64
import unittest
from unittest.mock import MagicMock, patch

from nerve.tools.compiler import wrap_tool_function


class TestWrapToolFunction(unittest.IsolatedAsyncioTestCase):
    @patch("nerve.tools.compiler.state")
    async def test_exception_in_wrapped_function_returns_error_string(self, _: MagicMock) -> None:
        def test_func() -> None:
            raise ValueError("test error")

        wrapped_func = wrap_tool_function(test_func)
        result = await wrapped_func()

        self.assertTrue(isinstance(result, str))
        self.assertTrue(result.startswith("ERROR in test_func: test error"))

    async def test_no_mime_returns_original_result(self) -> None:
        def test_func() -> str:
            return "test result"

        wrapped_func = wrap_tool_function(test_func)
        result = await wrapped_func()

        self.assertEqual(result, "test result")

    async def test_image_mime_returns_image_dict(self) -> None:
        test_bytes = b"test image data"
        expected_b64 = base64.b64encode(test_bytes).decode("utf-8")

        def test_func() -> bytes:
            return test_bytes

        mime = "image/png"
        wrapped_func = wrap_tool_function(test_func, mime=mime)
        result = await wrapped_func()

        self.assertEqual(result["type"], "image_url")
        self.assertEqual(result["image_url"]["url"], f"data:{mime};base64,{expected_b64}")

    @patch("nerve.tools.compiler.logger")
    @patch("nerve.tools.compiler.exit")
    async def test_invalid_mime_raises_exception(self, mock_exit: MagicMock, mock_logger: MagicMock) -> None:
        def test_func() -> str:
            return "test result"

        wrapped_func = wrap_tool_function(test_func, mime="invalid/mime")

        await wrapped_func()

        mock_logger.error.assert_called_once_with("tool test_func references an unsupported mime type: invalid/mime")
        mock_exit.assert_called_once_with(1)



import tempfile
import unittest
from pathlib import Path

from nerve.tools.namespaces import shell


class TestShell(unittest.TestCase):
    def test_shell(self) -> None:
        # Test basic command execution
        result = shell.shell("echo 'hello world'")
        self.assertEqual(result.strip(), "hello world")

    def test_shell_with_pipes(self) -> None:
        # Test command with pipes
        result = shell.shell("echo 'test' | tr 'e' 'E'")
        self.assertEqual(result.strip(), "tEst")

    def test_shell_with_environment_variables(self) -> None:
        # Test command with environment variables
        result = shell.shell("TEST_VAR='value' && echo $TEST_VAR")
        self.assertEqual(result.strip(), "value")

    def test_shell_with_file_operations(self) -> None:
        # Test command with file operations
        with tempfile.TemporaryDirectory() as temp_dir:
            test_file = Path(temp_dir) / "test.txt"

            # Create a file and write to it
            shell.shell(f"echo 'test content' > {test_file}")

            # Read the file
            self.assertTrue(test_file.exists())
            self.assertEqual(test_file.read_text().strip(), "test content")

    def test_shell_error(self) -> None:
        app = "foobarbazbiz666"
        output = shell.shell(app)
        self.assertTrue("EXIT CODE: " in output and "ERROR: " in output and "not found" in output and app in output)

    def test_shell_binary_output(self) -> None:
        result = shell.shell("head /bin/sh")
        # verify the result is bytes, not str
        self.assertIsInstance(result, bytes)
        self.assertTrue(result[:4] == b"\x7fELF" or result[:4] == b"\xca\xfe\xba\xbe")
